#include "XScene.h"
#include "XAcel.h"
#include "XStream.h"

#include <iostream>

using std::shared_ptr;

#pragma warning(  once : 4244  )

#include <scene-persistence/autogen/scenedb.pb.h>
#include <scene-persistence/autogen/stream_fsio.pb.h>

size_t XScene::GetStreamCount()
{
  return scene->streams_size();
}

PStream XScene::GetStream( size_t index )
{
  return PStream(new XStream(*scene->mutable_streams(index), work_dir ));
}

void XScene::InitializeNamedStreams()
{
  // build a names map object; reset names_streams to point to it
  named_streams.reset(new names_map_t());

  for (int idx = 0; idx < scene->streams_size(); ++idx)
  {
    const pb::Stream& stream = scene->streams(idx);
    if (stream.HasExtension(fsio::name))
      (*named_streams)[stream.GetExtension(fsio::name)] = idx;
  }
}

PStream XScene::GetStream( std::string name )
{
  // the named_streams map is built iff named streams are required
  if (named_streams == nullptr) InitializeNamedStreams();

  try { return this->GetStream(named_streams->at(name)); }

  catch (const std::out_of_range& oor) 
  {
    // std::cout << "OOR exception caught\n";
    return PStream(nullptr);
  }
  // std::map::at(const key_type& ) throws an out_of_range exception.
  // Instead of throwing an exception, we output a nullptr
}

PStream XScene::AddStream()
{
  size_t stream_index = scene->streams_size();
  scene->add_streams();
  return this->GetStream(stream_index);
}

PStream XScene::AddStream(std::string name)
{
  // check if stream already exists
  if (GetStream(name) != nullptr) throw ExistingStreamError(name);

  // add stream to protobuf scene
  size_t stream_index = scene->streams_size();
  scene->add_streams();
  // set name to newly added protobuf stream
  scene->mutable_streams(stream_index)->SetExtension(fsio::name, name);
  // add stream to index
  (*named_streams)[name] = stream_index;
  // return stream index
  return this->GetStream(stream_index);
}

bool XScene::DeleteStream(size_t index)
{
  if ((int)index >= scene->streams_size()) return false;

  // delete if from name list
  if (named_streams != nullptr)
  {
    pb::Stream* stream = scene->mutable_streams(index);
    if (stream->HasExtension(fsio::name))
      named_streams->erase(stream->GetExtension(fsio::name));
  }
  scene->mutable_streams()->DeleteSubrange(index,1);
  return true;
}

bool XScene::DeleteStream(std::string name)
{
  // the named_streams map is built iff named streams are required
  if (named_streams == nullptr) InitializeNamedStreams();

  int index;
  try { index = named_streams->at(name); }
  catch (std::out_of_range) { return false; }

  named_streams->erase(name);
  scene->mutable_streams()->DeleteSubrange(index,1);
  return true;
}

PAcel XScene::GetGlobals()
{
  return PAcel(XAcel::CreateAcel(*scene->mutable_globals()));
}

XScene::XScene( shared_ptr<pb::Scene> scene, std::string work_dir ) :
  scene(scene), work_dir(work_dir), named_streams(nullptr) 
{
  scene->mutable_globals()->set_id(1);
}

std::string XScene::GetRootDirectory()
{
  return work_dir;
}
