#pragma once

#ifndef SWIG
#include <string>
#include <memory>
#include <stdint.h>
#include <stdexcept>
#endif

#include <scene-persistence/autogen/scenedb.pb.h>

using std::string;

namespace sra
{

#pragma region forward declarations
  typedef int64_t oid_t;
  // forward declaration of the scene interfaces
  // sra data access
  class IScene;
  class IStream;
  class IAcel;
  class IAcelValue;
  class IAcelList;
#pragma endregion

  /// entry point to the scene implementation
  extern std::shared_ptr<IScene> get_scene_representation(string file);

  /// class representing scene container
  class IScene
  {
  public:
    /// returns acels defined in for all frames
    virtual std::shared_ptr<IAcel> GetGlobals() = 0;

    // get number of available streams
    virtual size_t GetStreamCount() = 0;
    // get new stream with index i
    virtual std::shared_ptr<IStream> GetStream(size_t index) = 0;
    // get stream from name
    virtual std::shared_ptr<IStream> GetStream(string name ) = 0;
    // adds new stream
    virtual std::shared_ptr<IStream> AddStream() = 0;
    // adds new stream with name
    virtual std::shared_ptr<IStream> AddStream(string name) = 0;
    // deletes a stream by index ~ does not delete the physical data!
    virtual bool DeleteStream(size_t index) = 0;
    // deletes a stream by name ~ does not delete the physical data!
    virtual bool DeleteStream(string name) = 0;

    // get root directory where scene file is stored. Asset paths are relative to this directory.
    virtual std::string GetRootDirectory() = 0;

    virtual ~IScene() {}
  };

  /// class representing sequence of generic frames
  class IStream
  {
  public:
    // sets stream data. Use when adding a new stream to a scene
    virtual void Initialize(std::shared_ptr<IAcel> stream_data)=0;
    // sets XStream to initial frame
    virtual void Reset()=0;
    // load next generic stream frame, return false if end of stream,
    // affects the frame that is returned by GetCurrentFrame
    virtual bool AdvanceToNextFrame()= 0;
    // get acel representing currently loaded frame
    virtual std::shared_ptr<IAcel> GetCurrentFrame()= 0;
    // get acel representing currently loaded frame
    virtual void SetCurrentFrame(std::shared_ptr<IAcel>)= 0;
    // get acel representing camera parameters used to capture this stream
    virtual std::shared_ptr<IAcel> GetCamera() = 0;
    // get acel for stream global metadata
    virtual std::shared_ptr<IAcel> GetStreamGlobals() = 0;

    // gets name of stream
    virtual std::string GetName() = 0;
    // gets path of the stream
    virtual std::string GetStreamPath() = 0;
    // get frame indexing information (-1 if implementation does not support it)
    virtual int GetFirstFrameIndex() = 0;
    virtual int GetLastFrameIndex() = 0;
    virtual int GetFrameIndexStep() = 0;

    virtual ~IStream() {}
  };

  // atomic scene element, containing pretty much any kind of hierarchical data
  class IAcel
  {
  public:
    // the unique id, equal for the acels with identical content
    virtual oid_t Id()=0;
    // set acel id. Return false if it has been defined.
    virtual bool SetId(oid_t id) = 0;
    // list of child acels
    // the list is owned by the acel
    // if there is no list, null pointer is returned
    virtual IAcelList* List() = 0;

    //  named child acels, if acel with this name doesn't exist UnsupportedProperty is thrown
    virtual std::shared_ptr<IAcel> Property(string name) = 0;
    // add named child acel with the unique id
    virtual std::shared_ptr<IAcel> AddProperty(string name, oid_t id) = 0;
    // add named child acel with the unique id. In this case, the acel is created according to a predefined protobuf message.
    // An object of autogenerated pb::Acel has to be allocated beforehand and a field with desired message type has to be set.
    virtual std::shared_ptr<IAcel> AddProperty(string name, oid_t id, pb::Acel* a) = 0;
    // delete named child acel
    virtual bool DeleteProperty(string name) = 0;

    // add child acel with the unique id
    virtual std::shared_ptr<IAcel> AddChild(oid_t id) = 0;
    // add child acel with the unique id. See AddProperty(string name, oid_t id, pb::Acel* a).
    virtual std::shared_ptr<IAcel> AddChild(oid_t id, pb::Acel* a) = 0;
    virtual bool DeleteChildId(oid_t id) = 0;
    virtual bool DeleteChild(size_t index) = 0;

    // value object, that can be converted to primitive data types
    // object is owned by the acel
    virtual IAcelValue* Value() = 0;
    virtual IAcelValue* SetValue() = 0;
    virtual void ClearValue() = 0;
    virtual ~IAcel() {}
  };

  struct UnsupportedPropertyError : public std::runtime_error
  { UnsupportedPropertyError(const std::string &d = "") : std::runtime_error(d) {}; };

  struct ExistingPropertyError : public std::runtime_error
  { ExistingPropertyError(const std::string &d = "") : std::runtime_error(d) {}; };

  struct InvalidAcelError : public std::runtime_error
  { InvalidAcelError(const std::string &d = "") : std::runtime_error(d) {}; };

  struct ExistingStreamError : public std::runtime_error
  { ExistingStreamError(const std::string &d = "") : std::runtime_error(d) {}; };

  // a collection of acels
  class IAcelList
  {
  public:
    /// number of acels in the list
    virtual size_t Count() const = 0;
    /// get acel from the list by index
    virtual std::shared_ptr<IAcel> Get(size_t index) = 0;
    virtual std::shared_ptr<IAcel> GetId(oid_t id) = 0;
    virtual ~IAcelList() {}
  };

#pragma region conversion of acel value to primitive types
  /// buffer contains a "Size" number of elements of type T
  /// memory is read-only and is owned by the Acel object
  template <typename T>
    class buffer_t
    {
      const T* ptr;
      size_t size;
    public:
      size_t Size() const { return size; }
      const T* Ptr() const { return ptr; }
      buffer_t(size_t n, const T* p): ptr(p), size(n) {}
    };
  /// primitive types
  using std::string;
  struct float3_t
  {
    float  x,y,z;
    float3_t(float x, float y, float z) : x(x), y(y), z(z) { }
  };

  // buffer is a pair, containing a pointer to the allocated c array of elements and the number of elements
  typedef buffer_t<uint8_t> byte_buffer_t;
  typedef buffer_t<float> float_buffer_t;
  typedef buffer_t<int32_t> int_buffer_t;

  /// this interface provides conversion of the acel into primitive data types
  /// the specific acel might implement one or more of these conversion methods
  /// others might throw an exception of type UnsupportedTypeError
  class IAcelValue
  {
  public:

    // scalars
    virtual float GetFloat() = 0;
    virtual void SetFloat(float val) = 0;
    virtual int32_t GetInt() = 0;
    virtual void SetInt(int32_t val) = 0;
    virtual string GetString() = 0;
    virtual void SetString(string val) = 0;
    virtual bool GetBool() = 0;
    virtual void SetBool(bool val) = 0;

    // vectors
    virtual float3_t GetFloat3() = 0;
    virtual void SetFloat3(float3_t val) = 0;

    // arrays of common basic types
    virtual byte_buffer_t GetByteBuffer()= 0;
    virtual void SetByteBuffer(byte_buffer_t val) = 0;

    virtual float_buffer_t GetFloatBuffer()= 0;
    virtual void SetFloatBuffer(float_buffer_t val) = 0;

    virtual int_buffer_t GetIntBuffer() = 0;
    virtual void SetIntBuffer(int_buffer_t val) = 0;

    virtual ~IAcelValue() {}
  };
  struct UnsupportedTypeError : public std::exception {};
#pragma  endregion

}

typedef std::shared_ptr<sra::IAcel> PAcel;
typedef std::shared_ptr<sra::IStream> PStream;
typedef std::shared_ptr<sra::IScene> PScene;

