﻿/*
 * backendCameraProjectionMatrix.cpp
 *
 *  Created on: Jun 5, 2014
 *      Author: mreinhardt
 */

// this repository
#include "scene-core/sra-api/sra_scene.h"
#include "scene-persistence/backend/backendCameraProjectionMatrix.h"

// external libraries

// operating system
#include <iostream>
#include <sstream>

static const std::string NAME   = "name";
static const std::string MATRIX = "matrix";
static const std::string STREAM = "stream";

#define DISABLE_EMPTY_STREAM_WARNING

int addCameraProjectionMatrix(PAcel parent,
                              IdManager::IdManager& idm,
                              const std::string cameraName,
                              const sra::float_buffer_t& projectionMatrix,
                              const std::vector<std::string>& correspondingStreams,
                              const int parentId)
{
	// Error checks
	if(!parent)
	{
		throw sra::InvalidAcelError("addCameraProjectionMatrix: Parent node is NULL.");
	}
	if(true == idm.MarkAsUsed(parentId))
	{
		std::cout << "Warning: The parentId was not yet marked as used. Fixing." << std::endl;
	}
	if(true == cameraName.empty())
	{
		throw sra::InvalidAcelError("addCameraProjectionMatrix: Cannot create a camera with an empty name.");
	}
	if(12 != projectionMatrix.Size() || NULL == projectionMatrix.Ptr())
	{
		// C++11 to_string() is not yet fully supported on all platforms.
		std::stringstream errorMsg;
		errorMsg << "addCameraProjectionMatrix: Projection matrix should have 12 valid entries. Given matrix has ";
		errorMsg << projectionMatrix.Size();
		errorMsg << " entries. Starting pointer is: ";
		errorMsg << projectionMatrix.Ptr();
		throw sra::InvalidAcelError(errorMsg.str().c_str());
	}
#ifndef DISABLE_EMPTY_STREAM_WARNING
	if(true == correspondingStreams.empty())
	{
		std::cout << "Warning: There are no corresponding stream names." << std::endl;
	}
#endif

	// Debug
// 	std::cout << "Name: " << cameraName << std::endl;
// 	std::cout << "Matrix: ";
// 	for (unsigned int index = 0; index < projectionMatrix.Size(); ++index)
// 	{
// 		std::cout << projectionMatrix.Ptr()[index] << ", ";
// 	}
// 	std::cout << std::endl;
// 	std::cout << "Streams: ";
// 	for(auto itStreams = correspondingStreams.begin(); itStreams != correspondingStreams.end(); ++itStreams)
// 	{
// 		std::cout << *itStreams << std::endl;
// 	}
// 	std::cout << "Parent Id: " << parentId << std::endl;
// 	std::cout << "Id of Acel: " << parent->Id() << std::endl;

	// Create camera acel
	PAcel cameraAcel = parent->AddChild(idm.AllocateId());
	PAcel nameProperty = cameraAcel->AddProperty(NAME, idm.AllocateId());
	nameProperty->SetValue();
	nameProperty->Value()->SetString(cameraName);

	PAcel matrixProperty = cameraAcel->AddProperty(MATRIX, idm.AllocateId());
	matrixProperty->SetValue();
	matrixProperty->Value()->SetFloatBuffer(projectionMatrix);

	if(false == correspondingStreams.empty())
	{
		PAcel streamProperty = cameraAcel->AddProperty(STREAM, idm.AllocateId());
		for(std::vector<std::string>::const_iterator itStreamNames = correspondingStreams.begin();
			itStreamNames != correspondingStreams.end(); ++itStreamNames)
		{
			PAcel someStreamName = streamProperty->AddChild(idm.AllocateId());
			someStreamName->SetValue();
			someStreamName->Value()->SetString(*itStreamNames);
		}
	}

	return parentId;
}

bool getCameraProjectionMatrix(PAcel parent,
                               std::string& cameraName,
                               sra::float_buffer_t& projectionMatrix,
                               std::vector<std::string>& correspondingStreams)
{
	if(!parent)
	{
		throw sra::InvalidAcelError("getCameraProjectionMatrix: Parent node is NULL.");
	}

	// get name
	try
	{
		PAcel cNameAcel = parent->Property(NAME);
		cameraName = cNameAcel->Value()->GetString();
	}
	catch(const sra::UnsupportedPropertyError& e)
	{
		std::cerr << "getCameraProjectionMatrix: Parsing acel failed. There is no \"" << NAME << "\" property in the given acel (id: " <<
			parent->Id() << ")." << std::endl;
		return false;
	}
	catch(const sra::UnsupportedTypeError& e)
	{
		std::cerr << "getCameraProjectionMatrix: Parsing acel failed. There is no value inside the \"" << NAME << "\" property of the given acel (id: " <<
			parent->Id() << ")." << std::endl;
		return false;
	}

	// get matrix
	try
	{
		PAcel cMatrixAcel = parent->Property(MATRIX);
		projectionMatrix = cMatrixAcel->Value()->GetFloatBuffer();
	}
	catch(const sra::UnsupportedPropertyError& e)
	{
		std::cerr << "getCameraProjectionMatrix: Parsing acel failed. There is no \"" << MATRIX << "\" property in the given acel (id: " <<
			parent->Id() << ")." << std::endl;
		return false;
	}
	catch(const sra::UnsupportedTypeError& e)
	{
		std::cerr << "getCameraProjectionMatrix: Parsing acel failed. There is no value inside the \"" << MATRIX << "\" property of the given acel (id: " <<
			parent->Id() << ")." << std::endl;
		return false;
	}

	// get streams
	correspondingStreams.clear();
	try
	{
		PAcel cStreamAcel = parent->Property(STREAM);
		sra::IAcelList* pStreamAcelList = cStreamAcel->List();
		if(NULL != pStreamAcelList)
		{
			for (size_t index = 0; index < pStreamAcelList->Count(); ++index)
			{
				correspondingStreams.push_back(pStreamAcelList->Get(index)->Value()->GetString());
			}
		}
	}
	catch(const sra::UnsupportedPropertyError& e)
	{
#ifndef DISABLE_EMPTY_STREAM_WARNING
		std::cout << "getCameraProjectionMatrix: Warning: There is no \"" << STREAM << "\" property in the given acel (id: " << parent->Id() << ")." << std::endl;
#endif
	}
	catch(const sra::UnsupportedTypeError& e)
	{
		std::cerr << "getCameraProjectionMatrix: Parsing acel failed. There is no value for at least on of the \"" << STREAM <<
			"\" property children of the given acel (id: " << parent->Id() << ")." << std::endl;
		return false;
	}

	return true;
}

PAcel getCameraByName(const PAcel parent, const std::string& camera_name)
{
	std::string name;
	sra::float_buffer_t matrix(0, NULL);
	std::vector<std::string> streams;

	sra::IAcelList * acels = parent->List();
	for (size_t i = 0; i < acels->Count(); ++i)
	{
		PAcel acel = acels->Get(i);
		if (::getCameraProjectionMatrix(acel, name, matrix, streams))
			if (name.compare(camera_name) == 0 )
				return acel;
	}

	return NULL;
}

std::vector<PAcel> getCamerasAssociatedToStream(const PAcel parent, const std::string& stream_name)
{
	std::string name;
	sra::float_buffer_t matrix(0, NULL);
	std::vector<std::string> streams;

	std::vector<PAcel> cameras(0);

	sra::IAcelList * acels = parent->List();
	for (size_t i = 0; i < acels->Count(); ++i)
	{
		PAcel acel = acels->Get(i);
		if (::getCameraProjectionMatrix(acel, name, matrix, streams))
			if (std::find(streams.begin(), streams.end(), stream_name) != streams.end())
				cameras.push_back(acel);
	}

	return cameras;
}

