﻿/*
 * backendCameraProjectionMatrix.h
 *
 *  Created on: Jun 5, 2014
 *      Author: mreinhardt
 */

#ifndef BACKENDCAMERAPROJECTIONMATRIX_H__
#define BACKENDCAMERAPROJECTIONMATRIX_H__

// this repository
#include "scene-core/sra-api/sra_scene.h"
#include "scene-persistence/backend/IdManager.h"

// external libraries

// operating system
#include <string>
#include <vector>


/**
 * Structure of a camera projection matrix acel.
 *
 *  Parent Node
 *     |
 *     |── properties
 *     |       |── name = [NAME]
 *     |       └── value ── value ── string_value =  [cameraName]
 *     |── properties
 *     |       |── name = [MATRIX]
 *     |       └── value ── value ── float_value_list =  [projectionMatrix]
 *     └── properties
 *             |── name = [STREAM]
 *             └── value ─┬─ children ── value ── string_value =  [correspondingStreams[0]]
 *                        ├─ children ── value ── string_value =  [correspondingStreams[1]]
 *                        ├─ children ── value ── string_value =  [correspondingStreams[2]]
 *                        ...
 *
 * NOTE: The correspondingStreams parameter is optional. There will be no Property named "stream" when there is
 *       corresponding stream. In other words, if there is such a property at least one stream named must be given.
 */

/**
 * Create an acel containing the camera projection matrix and some identification information. The acel will be a
 * direct child of the given node.
 *
 * @param[in]  parent                 Node to which the camera acel shall be added.
 * @param[in]  idm                    IdManager to generate new Ids. parentId should already be marked as used.
 * @param[in]  cameraName             Name of the camera. This can be any type of string. Empty strings are not allowed.
 * @param[in]  projectionMatrix       3x4 projection matrix for this camera. It is recommended to store the values row-wise.
 * @param[in]  correspondingStreams   List of names of streams this camera is used for. This parameter is optional
 * @param[in]  parentId               Id of the camera acel that shall be created. Should be marked as set in idm.
 *
 * @throws InvalidAcelError  The Parameter parent is not a valid pointer.
 * @throws InvalidAcelError  The Parameter parent is not a valid pointer.
 *
 * @return Id of the camera acel that was created. Identical to parentId.
 */
int addCameraProjectionMatrix(PAcel parent,
                              IdManager::IdManager &idm,
                              const std::string cameraName,
                              const sra::float_buffer_t &projectionMatrix,
                              const std::vector<std::string> &correspondingStreams,
                              const int parentId);

/**
 * Parse an acel containing the camera projection matrix and some identification information. The up to three properties
 * of the camera acel are expected to be direct members of the given node.
 *
 * @param[in]  cameraAcel           Node that contains the camera properties. Should have the Id that was returned
 *                                  by addCameraProjectionMatrix().
 * @param[out] cameraName           Name that was found in the NAME property of the parent acel.
 * @param[out] projectionMatrix     Projection matrix that was found in the MATRIX property of the parent acel.
 * @param[out] correspondingStreams List of names of streams this camera is used for. Found in the STREAM property of the
 *                                  parent acel. May be empty.
 *
 * @return True == Node was parsed successfully.
 *         False == Parsing of the node failed. Using the return values will result in undefined behavior.
 */
bool getCameraProjectionMatrix(PAcel cameraAcel,
                               std::string &cameraName,
                               sra::float_buffer_t &projectionMatrix,
                               std::vector<std::string> &correspondingStreams );

/**
 * Searches a camera projection matrix Acel by name among the children of a
 * specified parent Acel. 
 *
 * @param[in]  parent               The camera projection matrix is going to be
 *                                  search for among the children of this node.
 * @param[in]  cameraName           Query camera name.
 *
 * @return The camera projection matrix Acel found with the specified name. 
 *         If no camera projection matrix was found, returns NULL. If the
 *         parent Acel has more than one camera projection matrices it returns
 *         the first in the children list.
 */
PAcel getCameraByName(const PAcel parent, const std::string& cameraName);

/**
 * Retrieves all camera projection matrix Acels associated with a specified
 * stream, among the children of a specified parent Acel. If none were
 * found, it returns an empty std::vector. 
 *
 * @param[in]  parent               The camera projection matrix is going to be
 *                                  search for among the children of this node.
 * @param[in]  streamName           Stream name.
 *
 * @return An std::vector with the camera projection matrix Acels under
 *         'parent', which are associated with the specified stream. If no
 *         camera projection matrix was found, returns an empty vector.
 */
std::vector<PAcel> getCamerasAssociatedToStream(const PAcel parent, const std::string& streamName);


#endif // BACKENDCAMERAPROJECTIONMATRIX_H__
