/*
 * sraBackend.cpp
 *
 *  Created on: Mar 4, 2014
 *      Author: kelvinchelli
 */

#include "sraBackend.h"
#include "backendCameraProjectionMatrix.h"

#include <sstream>

namespace backend {

/*
 * function to add an image to a Scene
 */
int set_Map(PScene scene, IdManager::IdManager &idm, std::string filename, mapType::type type, int parentId) {
	if(idm.MarkAsUsed(parentId)) { 	// check if parentId is unused
		idm.FreeId(parentId);
		parentId = 0;			// if it is, set parentId to 0 (top layer)
	}
	int id = idm.AllocateId();
	if(parentId==0) { 	// check if parent ID 0
		setMap(scene->GetGlobals(),filename,idm,id,type);
	} else {			// Acel with parentId exists
		setMap(getAcel(scene->GetGlobals(),parentId),filename,idm,id,type);
	}
	cout << "map id is " << id << endl;
	return id;
}


/*
 * function to add a mesh to a Scene
 */
int set_Mesh(PScene scene, IdManager::IdManager &idM, std::string filename, meshType::type type, int parentId, int **ids) {
	int numOfMeshes = 0;
	if(idM.MarkAsUsed(parentId)) { 	// check if parentId is unused
		idM.FreeId(parentId);
		parentId = 0;				// if it is, set parentId to 0 (top layer)
	}
	if(parentId==0) { 	// check if parent ID 0
		numOfMeshes = setMesh(scene->GetGlobals(), filename, idM, type, ids);
	} else {			// Acel with parentId exists
		numOfMeshes = setMesh(getAcel(scene->GetGlobals(),parentId), filename, idM, type, ids);
	}

	return numOfMeshes;
}

/*
 * get path_to_file from acel with id
 */
string getFilePath(PScene scene, int id) {
	PAcel acel = getAcel(scene->GetGlobals(),id);
	if(acel->Property("path_to_file")) {
		// if found, return file path
		return acel->Property("path_to_file")->Value()->GetString();
	}
	cerr << "Warning: Property path_to_file not set for Acel " << id << "." << endl;
	return "";
}

/*
 * write file path to acel
 */
int setFilePath(PScene scene, IdManager::IdManager &idM, int ParentId, string filename) {
	if(idM.MarkAsUsed(ParentId)) { 	// check if parentId is unused
		idM.FreeId(ParentId);
		ParentId = 0;				// if it is, set parentId to 0 (top layer)
	} //else if(getAcel(scene->GetGlobals(),ParentId)->???()){					// check if acel with ID is a property
		//TODO: catch case when acel is property: cannot have child acels
	//}
	int id = idM.AllocateId();
	if(ParentId==0) { 	// check if parent ID 0
		PAcel parent = scene->GetGlobals()->AddChild(id);
		PAcel path = parent->AddProperty("path_to_file",idM.AllocateId());
		path->SetValue()->SetString(filename);
	} else {			// Acel with parentId exists
		PAcel parent = getAcel(scene->GetGlobals(), ParentId)->AddChild(id);
		PAcel path = parent->AddProperty("path_to_file",idM.AllocateId());
		path->SetValue()->SetString(filename);
	}
	return id;
}

/*
 * get binary acel data from acel with id
 */
int getBinaryFile(PScene scene, int id, char **buffer) {
	int length = 0;
	PAcel acel = getAcel(scene->GetGlobals(),id);
	if(acel->Property("binary_data")) {
		// if found, allocate memory at buffer
		length = acel->Property("binary_data")->Value()->GetByteBuffer().Size();

		*buffer = new char[length];

		for(int i=0; i<length; i++) {
			*(*buffer+i) = acel->Property("binary_data")->Value()->GetByteBuffer().Ptr()[i];
		}

		return length;
	}
	cerr << "Warning: Property binary_data not set for Acel " << id << "." << endl;
	return length;
}

/*
 * write binary data to acel
 */
int setBinaryFile(PScene scene, IdManager::IdManager &idM, int ParentId, string filename){
	std::ifstream infile (filename, std::ifstream::binary);
	size_t size = 0; // here
	int id = 0;
	if (infile) {
		// get length of file:
		infile.seekg (0, infile.end);
		int length = infile.tellg();
		infile.seekg (0, infile.beg);

		char * buffer = new char [length];

		//std::cout << "Reading " << length << " characters... ";
		// read data as a block:
		infile.read (buffer,length);

		if (infile) {}
			//std::cout << "all characters read successfully.";
		else {
			std::cout << "error: only " << infile.gcount() << " could be read"; }
		infile.close();

	    // ...buffer contains the entire file...
		if(idM.MarkAsUsed(ParentId)) { 	// check if parentId is unused
			idM.FreeId(ParentId);
			ParentId = 0;				// if it is, set parentId to 0 (top layer)
		} //else if(getAcel(scene->GetGlobals(),ParentId)->???()){					// check if acel with ID is a property
			//TODO: catch case when acel is property: cannot have child acels
		//}
		id = idM.AllocateId();
		PAcel parent;
		if(ParentId==0) { 	// check if parent ID 0
			parent = scene->GetGlobals()->AddChild(id);
		} else {			// Acel with parentId exists
			parent = getAcel(scene->GetGlobals(), ParentId)->AddChild(id);
		}
		PAcel path = parent->AddProperty("binary_data",idM.AllocateId());
		path->SetValue()->SetByteBuffer(sra::byte_buffer_t(size,(uint8_t*)buffer));

	    delete[] buffer;
	}
	else {
		cout << "File " << filename << " does not exist." << endl;
		throw std::invalid_argument(filename);
	}

	return id;
}

/**
 * find Acel with certain Id as child Acel of parent and return reference this acel
 * return null if no Acel with ID exists
 */
PAcel getAcel(PAcel parent, int id) {
	sra::IAcelList* children = parent->List();
	for(size_t i=0; i<children->Count(); ++i) {
		if(children->Get(i)->Id()==id) {
			return children->Get(i);
		}
	}

	for (size_t i=0; i < children->Count(); ++i) {
		PAcel ret = getAcel(children->Get(i),id);
		if(NULL!=ret){
			return getAcel(children->Get(i),id);
		}
	}

	return NULL;
}

int addCameraProjectionMatrix(PScene scene,
                              IdManager::IdManager &idm,
                              const std::string cameraName,
                              const sra::float_buffer_t &projectionMatrix,
                              const std::vector<std::string> &correspondingStreams,
                              const int parentId)
{
	if(!scene)
	{
		throw sra::InvalidAcelError("addCameraProjectionMatrix: No scene graph given.");
	}
	if(true == idm.MarkAsUsed(parentId)) { 	// check if parentId is unused
		idm.FreeId(parentId);
		std::stringstream errorMsg;
		errorMsg << "addCameraProjectionMatrix: cannot add camera acel to non-existing acel with id";
		errorMsg << parentId;
		throw sra::InvalidAcelError(errorMsg.str().c_str());
	}
	int id = idm.AllocateId();
	if(1 == parentId) { 	// check if parent ID 0
		::addCameraProjectionMatrix(scene->GetGlobals(), idm, cameraName, projectionMatrix, correspondingStreams, parentId);
	} else {			// Acel with parentId exists
		::addCameraProjectionMatrix(getAcel(scene->GetGlobals(), parentId), idm, cameraName, projectionMatrix, correspondingStreams, parentId);
	}
	cout << "camera projection id is " << id << endl;
	return id;
}

bool getCameraProjectionMatrix(const PScene scene,
                               const int id,
                               std::string &cameraName,
                               sra::float_buffer_t &projectionMatrix,
                               std::vector<std::string> &correspondingStreams)
{
	if(!scene)
	{
		throw sra::InvalidAcelError("addCameraProjectionMatrix: No scene graph given.");
	}
	if(1 == id) { 	// check if parent ID 0
		return ::getCameraProjectionMatrix(scene->GetGlobals(), cameraName, projectionMatrix, correspondingStreams);
	} else {			// Acel with parent ID exists
		return ::getCameraProjectionMatrix(getAcel(scene->GetGlobals(), id), cameraName, projectionMatrix, correspondingStreams);
	}
}

/*
 * Id Management during initialization
 */
void blockIds(PScene scene, IdManager::IdManager &idM){
	// block already existing id's in scene file
	PAcel globals = scene->GetGlobals();
	idM.MarkAsUsed(globals->Id());
	blockChildIds(globals, idM);
}
void blockChildIds(PAcel acel, IdManager::IdManager &idM){
	// block ids of children in id manager
	sra::IAcelList* children = acel->List();
	for(size_t i=0; i<children->Count(); ++i) {
		idM.MarkAsUsed(children->Get(i)->Id());
	}

	// for all child acels also block ids in id manager
	for(size_t i=0; i<children->Count(); ++i) {
		PAcel child = children->Get(i);
		blockChildIds(child, idM);
	}
}


bool remove_Acel(PScene scene, IdManager::IdManager &idm, int id){
	int removed = 0;
	removed = removeChildAcel(scene->GetGlobals(), id);
	if(removed==2) {
		idm.FreeId(id);
		return true;
	}

	return false;
}
int removeChildAcel(PAcel acel, int id){
	int ret = 0;
	// check if acel has wanted ID
	if(acel->Id()==id) {
		return 1;
	}

	// for all child acels also allocate ids in id manager
	sra::IAcelList* children = acel->List();
	for (size_t i=0; i < children->Count(); i++) {
		ret = removeChildAcel(children->Get(i), id);
		if(ret==1){
			acel->DeleteChild(i);
			return 2;
		}
	}
	return 0;
}

} /* namespace backend */
