#include <cstdio>
#include <fcntl.h>

#include "scene-core/sra-api/sra_scene.h"
#include "scene-core/XScene.h"
#include "scene-core/sra_implementation.h"
#include "scene-persistence/autogen/scenedb.pb.h"
#include "scene-persistence/backend/IdManager.h"
#include "scene-persistence/backend/sraBackend.h"

#define PATH argv[1] // path to root directory of the SRA
#define NAME argv[2] // name of the output scene file
int main(int argc, const char* argv[])
{
	// help message
	if (argc != 3)
	{
		std::printf("Usage: %s root-sra-directory name-of-scene-file \n\n",argv[0]);
		std::printf("Generates an SRA to test the Digital Cinema workflow.\n");
		std::printf("The scene file is stored in the directory from where \n");
		std::printf("the program is called. It assumes the following folder\n");
		std::printf("structure:\n");
		std::printf("\n");
		std::printf("      root-sra-directory\n");
		std::printf("      	├── ALEXA_A\n");
		std::printf("      	│   ├── 363015.exr\n");
		std::printf("      	│   ├── 363016.exr\n");
		std::printf("      	│   ├── 363017.exr\n");
		std::printf("      	│   ...\n");
		std::printf("      	└── ALEXA_B\n");
		std::printf("      	    ├── 363015.exr\n");
		std::printf("      	    ├── 363016.exr\n");
		std::printf("      	    ├── 363017.exr\n");
		std::printf("      	    ...   \n\n");
		return EXIT_FAILURE;
	}

	// create empty scene
	PScene scene = sra::make_scene_representation("./");

	// acel id generator, starts from 2 because 1 is reserved for scene globals
	IdManager::IdManager idm;
	idm.MarkAsUsed(1);

	std::string sra_root_dir(PATH);

	// add rgb stream
	{
		// add an empty stream called "rgb"
		PStream stream = scene->AddStream("rgb");

		// newly added streams have to be initialized, to that aim, we create an
		// acel with the necessary stream metadata
		PAcel stream_data = scene->GetGlobals()->AddChild(idm.AllocateId());

		// path to the stream files (relative to scene working directory) - mandatory
		stream_data->AddProperty("video", idm.AllocateId())->SetValue()->SetString(sra_root_dir + "/ALEXA_A/%06d.exr");

		// first frame - optional: if no number is specified the stream path
		// is search for files. The first frame is set according to the first
		// frame found (this process can take time, thus we provide the first
		// frame).
		stream_data->AddProperty("first_frame", idm.AllocateId())->SetValue()->SetInt(363015);

		// stream name - optional. This property is usefull when the string is created
		// with the XScene::AddStream(void) method.
	//	stream_data->AddProperty("name", idm.AllocateId())->SetValue()->SetString("rgb");

		// step between consecutive frame numbers - optional
	//	stream_data->AddProperty("frame_step", idm.AllocateId())->SetValue()->SetInt(1);

		// last_frame - optional 
	//	stream_data->AddProperty("frame_step", idm.AllocateId())->SetValue()->SetInt(363018);

		// names of the channels in the EXR files - if no channels are specified, 
		// all the channels found in the EXR file are used (as long as they
		// all have the same image dimensions and type)
		stream_data->AddProperty("exr_channel", idm.AllocateId());
		stream_data->Property("exr_channel")->AddChild(idm.AllocateId())->SetValue()->SetString("B");
		stream_data->Property("exr_channel")->AddChild(idm.AllocateId())->SetValue()->SetString("G");
		stream_data->Property("exr_channel")->AddChild(idm.AllocateId())->SetValue()->SetString("R");

		stream->Initialize(stream_data);

		// we don't need stream_data acel anymore
		scene->GetGlobals()->DeleteChildId(stream_data->Id());
	}

	// add tof depth stream
	{
		PStream stream = scene->AddStream("tof_depth");

		PAcel stream_data = scene->GetGlobals()->AddChild(idm.AllocateId());
		stream_data->AddProperty("video", idm.AllocateId())->SetValue()->SetString(sra_root_dir + "/ALEXA_A/%06d.exr");
		stream_data->AddProperty("first_frame", idm.AllocateId())->SetValue()->SetInt(363015);
		stream_data->AddProperty("exr_channel", idm.AllocateId());
		stream_data->Property("exr_channel")->AddChild(idm.AllocateId())->SetValue()->SetString("depth.Z");

		stream->Initialize(stream_data);
		scene->GetGlobals()->DeleteChildId(stream_data->Id());
	}

	// add rgb stream
	{
		// add an empty stream called "rgb"
		PStream stream = scene->AddStream("satellite_rgb");

		PAcel stream_data = scene->GetGlobals()->AddChild(idm.AllocateId());
		stream_data->AddProperty("video", idm.AllocateId())->SetValue()->SetString(sra_root_dir + "/ALEXA_B/%06d.exr");
		stream_data->AddProperty("first_frame", idm.AllocateId())->SetValue()->SetInt(363015);
		stream_data->AddProperty("exr_channel", idm.AllocateId());
		stream_data->Property("exr_channel")->AddChild(idm.AllocateId())->SetValue()->SetString("B");
		stream_data->Property("exr_channel")->AddChild(idm.AllocateId())->SetValue()->SetString("G");
		stream_data->Property("exr_channel")->AddChild(idm.AllocateId())->SetValue()->SetString("R");

		stream->Initialize(stream_data);
		scene->GetGlobals()->DeleteChildId(stream_data->Id());
	}


	// add HHI's camera projection matrix
	// FIXME: currently camera data has to be hardcoded!
	{
		// array of 12 floats
		const float proj_matrix_data[12] = {11, 12, 13, 14,
		                                    21, 22, 23, 24,
		                                    31, 32, 33, 34 };

		// wrap it into the sra::float_buffer_t
		sra::float_buffer_t proj_matrix(12, proj_matrix_data);

		// optionally, the projection matrix can be associated with a vector of streams
		std::vector<std::string> associated_streams;
		associated_streams.push_back("rgb");
		associated_streams.push_back("tof_depth");

		// give a name to the camera
		string camera_name("ALEXA_A");

		// The camera data will be stored in an Acel, which itself, needs to be 
		// the children of another Acel exiting in the scene. We need to specify 
		// the id of the parent Acel to which we will "attach" the camera. In this
		// case, we will attach it to the top Acel of the scene, namely
		// scene->Globals(), which always has index 1.
		int parent_id = 1;
	
		backend::addCameraProjectionMatrix(scene, idm, camera_name, proj_matrix,
				associated_streams, parent_id);
	}

	/* TEST: add HHI's camera projection matrix
	{
		// array of 12 floats
		const float proj_matrix_data[12] = {1.1, 1.2, 1.3, 1.4,
		                                    2.1, 2.2, 2.3, 2.4,
		                                    3.1, 3.2, 3.3, 3.4 };

		// wrap it into the sra::float_buffer_t
		sra::float_buffer_t proj_matrix(12, proj_matrix_data);

		// optionally, the projection matrix can be associated with a vector of streams
		std::vector<std::string> associated_streams;
		associated_streams.push_back("satellite_rgb");
		associated_streams.push_back("rgb");

		// give a name to the camera
		string camera_name("ALEXA_A_2");

		// The camera data will be stored in an Acel, which itself, needs to be 
		// the children of another Acel exiting in the scene. We need to specify 
		// the id of the parent Acel to which we will "attach" the camera. In this
		// case, we will attach it to the top Acel of the scene, namely
		// scene->Globals(), which always has index 1.
		int parent_id = 1;
	
		backend::addCameraProjectionMatrix(scene, idm, camera_name, proj_matrix,
				associated_streams, parent_id);
	}*/

	
	// write the .scene file. If the third boolean parameter is true a 
	// human-readable description of the scene file is also generated 
	// in a file named by appending .txt to the scene filename. This
	// is mainly usefull for debugging.
	std::string scene_filename(NAME); scene_filename += ".scene";
	sra::set_scene_representation(scene_filename, scene, true);

	return EXIT_SUCCESS;
}
