#include "scene-core/sra-api/sra_scene.h"
#include "scene-core/XScene.h"
#include "scene-core/sra_implementation.h"
#include "scene-persistence/autogen/scenedb.pb.h"
#include "scene-persistence/backend/backendCameraProjectionMatrix.h"

#include <cstdio>
#include <fcntl.h>

#include <opencv2/opencv.hpp>

void printCameraProjectionMatrix(PAcel cam)
{
	std::string name;
	std::vector<std::string> streams;
	sra::float_buffer_t matrix(0, NULL);

	::getCameraProjectionMatrix(cam, name, matrix, streams);

	std::cout << "\tCamera name: " << name << std::endl; 

	std::cout << "\tMatrix data: " << std::endl; 
	printf("\t\t%f %f %f %f\n \t\t%f %f %f %f\n \t\t%f %f %f %f\n",
			matrix.Ptr()[ 0], matrix.Ptr()[ 1], matrix.Ptr()[ 2], matrix.Ptr()[ 3],
			matrix.Ptr()[ 4], matrix.Ptr()[ 5], matrix.Ptr()[ 6], matrix.Ptr()[ 7],
			matrix.Ptr()[ 8], matrix.Ptr()[ 9], matrix.Ptr()[10], matrix.Ptr()[11]);

	std::cout << "\tAssociated streams: \n";
	std::vector<std::string>::iterator stream_name; 
	for (stream_name = streams.begin(); stream_name != streams.end(); ++stream_name)
		std::cout << "\t\t" << *stream_name << std::endl;

	std::cout << "\n";
}


#define SRA argv[1] // path to SRA file
#define STR argv[2] // stream name
#define OUT argv[3] // output folder
int main(int argc, const char* argv[])
{
	// help message
	if (argc != 4)
	{
		std::printf("Usage: %s sra-file stream-name output-folder\n\n",argv[0]);
		std::printf("Reads a single stream from a multistream SRA and\n");
		std::printf("writes it as png files in the output folder.\n");
		return EXIT_FAILURE;
	}

	// load scene filename
	std::string scene_filename(SRA);
	PScene scene = sra::get_scene_representation(scene_filename);

	// read stream
	// get stream - returns NULLPTR if stream does not exist
	PStream stream = scene->GetStream(STR);

	if (stream)
	{
		// retrieve each frame, and save it as png images
		while (stream->AdvanceToNextFrame()) 
		{
			PAcel frame_acel = stream->GetCurrentFrame();

			// retrieve frame info from frame_acel
			int w = frame_acel->Property(XSTREAM_INFO_VIDEO_WIDTH   )->Value()->GetInt(); // frame width
			int h = frame_acel->Property(XSTREAM_INFO_VIDEO_HEIGHT  )->Value()->GetInt(); // height
			int c = frame_acel->Property(XSTREAM_INFO_VIDEO_CHANNELS)->Value()->GetInt(); // channels
			int d = frame_acel->Property(XSTREAM_INFO_VIDEO_DEPTH   )->Value()->GetInt(); // pixel depth in bytes
			int f = frame_acel->Property(XSTREAM_INFO_VIDEO_INDEX   )->Value()->GetInt(); // frame number

			// pointer to the data
			const uint8_t* data = frame_acel->Property(XSTREAM_FRAME_VIDEO_DATA)->Value()->GetByteBuffer().Ptr();

			/* NOTE: the SRA only supports two OpenEXR pixel types:
			 *     UINT - unsigned 16 bit integer
			 *     HALF - 16 bit floating point
			 *
			 * EXR channels of HALF type are returned as floats (32 bits), so pixel depth
			 * is either 2 (for UINT) and 4 for HALF. */

			// wrap data in a opencv image and write to disk
			{
				int type = (d == 2) ? CV_MAKETYPE(CV_16U,c) : CV_MAKETYPE(CV_32F,c);
				cv::Mat frame(h, w, type, (void *)data);

				// if needed, convert to 16b unsiged char
				cv::Mat frame_16u;
				frame.convertTo(frame_16u, CV_16U, 256*256);

				char name[128];
				std::sprintf(name,"%s/%s_%06d.png",OUT,STR,f);
				cv::imwrite(name,frame_16u);
			}
		}

		// retrieve camera projection matrices which have the selected stream as 
		// associated stream and display information
		std::vector<PAcel> cameras = ::getCamerasAssociatedToStream(scene->GetGlobals(), STR);

		std::cout << "The following cameras associated to stream " << STR << " were found\n";

		std::vector<PAcel>::iterator cam;
		for(cam = cameras.begin(); cam != cameras.end(); ++cam)
			printCameraProjectionMatrix(*cam);
	}
	else std::cout << "No " << STR << " stream found\n";

	/* TEST: retrieve camera projection matrices by camera name
	PAcel camera = ::getCameraByName(scene->GetGlobals(), "ALEXA_A");

	std::cout << "The following camera named 'ALEXA_A' was found\n";
	printCameraProjectionMatrix(camera);*/

	return EXIT_SUCCESS;
}
