/*
 * sra.h
 *
 */

#ifndef SRA_H_
#define SRA_H_

#include <string>
#include <fstream>
#include <iostream>
#include <vector>

#include "backend/sraBackend.h"
#include "backend/filetypes.h"
#include <scene-core/sra_implementation.h>
#include <scene-core/sra-api/sra_scene.h>


using namespace cimg_library;

class Scene {
	PScene scene;
	std::string sceneFile;
	IdManager::IdManager idM;
public:
	Scene(std::string sceneFile);	// constructor with filename of scene file
	virtual ~Scene();			// default destructor

	int writeScene(std::string sceneFile);	// write Scene to file

	bool removeAcel(int id);	// remove Acel from scene

	int addMesh(std::string filename, meshType::type type, int parentId, int **ids);
	aiMesh *getMesh(int id);

	int addMap(std::string filename, mapType::type type, int parentId);
	CImg<unsigned char> *getMap(int id);

	int addFilePath(std::string filename, int parentId);
	std::string getFilePath(int id);

	int addBinaryData(std::string filename, int parentId);
	int getBinaryData(int id, char **buffer);

	/**
	 * Create an acel containing the camera projection matrix and some identification information. The acel will be a
	 * to the acel with the id parentId.
	 *
	 * @param[in]  cameraName             Name of the camera. This can be any type of string. Empty strings are not allowed.
	 * @param[in]  projectionMatrix       3x4 projection matrix for this camera. It is recommended to store the values row-wise.
	 * @param[in]  correspondingStreams   List of names of streams this camera is used for. This parameter is optional
	 * @param[in]  parentId               Id of the parent acel where the new camera acel will be attached to.
	 *
	 * @throws InvalidAcelError  Should not be thrown. Would indicate an error in the creation routine of this object.
	 *
	 * @return Id of the camera acel that was created.
	 */
	int addCameraProjectionMatrix(const std::string cameraName,
	                              const sra::float_buffer_t& projectionMatrix,
	                              const std::vector<std::string>& correspondingStreams,
	                              const int parentId);

	/**
	 * Get the information for a specific camera acel. The up to three properties of the camera acel are expected to be
	 * direct members of the given node.
	 *
	 * @param[in]  id                   Id of the parent acel of the camera acel. This is the return value of addCameraProjectionMatrix().
	 * @param[out] cameraName           Name that was found in the NAME property of the parent acel.
	 * @param[out] projectionMatrix     Projection matrix that was found in the MATRIX property of the parent acel.
	 * @param[out] correspondingStreams List of names of streams this camera is used for. Found in the STREAM property of the
	 *                                  parent acel. May be empty.
	 *
	 * @throws InvalidAcelError  The given scene graph is invalid.
	 *
	 * @return True == Node was parsed successfully.
	 *         False == Parsing of the node failed. Using the return values will result in undefined behavior.
	 */
	bool getCameraProjectionMatrix(const int id,
	                               std::string& cameraName,
	                               sra::float_buffer_t& projectionMatrix,
	                               std::vector<std::string>& correspondingStreams) const;

	/* Pablo: produces a compilation error, since DoF is not defined as a type
	 * elsewhere
	int addDoF(string filename, int parentId);
	DoF *getDoF(int id);
	*/

	PScene getScene();

private:
	//int createCamera();
	//int fakeCamera();
	//int convertTexture();
	//int createDepth();

	//PAcel getGlobals();

	//int convertGeometry();
};



#endif /* SRA_H_ */
