#include "exr_frames_texture_loader.h"
#include <stdexcept>
#include <iostream>
#include <fstream>
#include <cmath>

#include <scene-persistence/autogen/scenedb.pb.h>
#include <scene-persistence/autogen/stream_fsio.pb.h>

#include <opencv2/opencv.hpp> // store channels read from EXR image
#include "exr_utils/exr_utils.h" // read channels from EXR images

#include "frame_seq_utils.h"

bool EXRTextureLoader::NextFrame()
{
  current_frame += frame_step;

  if (current_frame < last_frame)
  {
    // form frame's filename
    char * name = new char[frames_name.size() + 10];
    std::sprintf(name,frames_name.c_str(),current_frame);

    // open image and set video_buffer
    try { exr_utils::readEXR_MultichannelInterleaved( name, channel_names, *frame ); }
    catch (...)
    {
      delete [] name;
      return false;
    }

    delete [] name;

    video_buffer = (uint8_t *)frame->data;

    // in case some problem occured, even if file exists
    if (!frame->data) return false;

	 if (frame->cols != width || frame->rows != height || frame->channels() != channels)
		 throw std::runtime_error("Loaded frame size/channels do not match EXRTextureLoader's");
	 if ((int)(frame->elemSize1()) != depth)
		 throw std::runtime_error("Loaded frame pixel depth do not match EXRTextureLoader's");

    return true;
  }
  
  // current_frame >= last_frame
  return false;
}


void EXRTextureLoader::SetCurrentData(const uint8_t *data, int w, int h, int c, int d)
{
  if (current_frame < first_frame)
	  throw std::runtime_error("current_frame < first_frame in EXRTextureLoader::SetCurrentData");

  // for a newly created file stream, dimensions are not initialized yet
  if (!(width > 0 && height > 0 && channels > 0 && depth > 0))
  {
    if (!(w > 0 && h > 0 && c > 0 && d > 0))
		 throw std::runtime_error("Calling EXRTextureLoader::SetCurrentData "
				 "without dimensions/channels/pixel-depth for a EXRTextureLoader "
				 "whose dimensions/channel/pixel-depth have not yet been initialized");
	
    width = w;
    height = h;
    channels = c;
    depth = d;
  }
  // if dimensions have already been initialized, input dimensions should match them
  else
  {
    if (w > 0 && w != width   ) throw std::runtime_error("Input width doesn't match EXRTextureLoader's width");
    if (h > 0 && h != height  ) throw std::runtime_error("Input height doesn't match EXRTextureLoader's height");
    if (c > 0 && c != channels) throw std::runtime_error("Input channels don't match EXRTextureLoader's channels");
    if (d > 0 && d != depth   ) throw std::runtime_error("Input pixel-depth doesn't match EXRTextureLoader's pixel-depth");
  }


  // form frame's filename
  char * name = new char[frames_name.size() + 10];
  std::sprintf(name,frames_name.c_str(),current_frame);
  
  // write image
  int type;
  switch (depth)
  {
    case 2: type = CV_MAKETYPE(CV_16U,channels); break;
    case 4: type = CV_MAKETYPE(CV_32F,channels); break;
	 default: throw std::runtime_error("Unsupported pixel-depth in EXRTextureLoader");
  }
  cv::Mat frame = cv::Mat(height, width, CV_MAKETYPE(type, channels), (void *)data);
  exr_utils::appendEXR_MultichannelInterleaved( name, channel_names, frame );
  delete [] name;

  // if current frame exceeds last_frame, update last_frame
  if (current_frame == last_frame)
  {
    last_frame += frame_step;
    if (stream) stream->SetExtension(fsio::last_frame, last_frame-1);
  }
}


void EXRTextureLoader::Initialize(void)
{
  // tries to set current_frame, width, height, channels, depth and
  // channel_names by reading the first file in the file sequence.
  // This function is used by the constructors only. 
 
  // set current frame
  current_frame = first_frame - 1;

  // if no channel names are specified, read them from first stream file
  if (channel_names.size() == 0)
  {
    // form frame's filename
    char * name = new char[frames_name.size() + 10];
    std::sprintf(name,frames_name.c_str(),first_frame);
    try
    {
      exr_utils::readEXR_MultichannelInterleaved( name, *frame , channel_names);
    }
    catch (...) {}
    delete [] name;
  }

  // open first frame to set height, width and channels
  {
    if (frame->data == nullptr)
    {
      // form frame's filename
      char * name = new char[frames_name.size() + 10];
      std::sprintf(name,frames_name.c_str(),first_frame);
//    std::cout << "\t\treading image " << name << "\n";
//    for (size_t ch = 0; ch < channel_names.size(); ++ch)
//      std::cout << "\t\t\t" << channel_names[ch] << "\n";

      try { exr_utils::readEXR_MultichannelInterleaved( name, channel_names, *frame ); }
      catch (...) { std::cout << "\t\tcould not open first image in stream\n"; }
      delete [] name;
    }

    height   = frame->data ? frame->rows        : -1;
    width    = frame->data ? frame->cols        : -1;
    channels = frame->data ? frame->channels()  : -1;
    depth    = frame->data ? frame->elemSize1() : -1;
  }
}

EXRTextureLoader::EXRTextureLoader( std::string filename, 
    int firstf, int lastf, int fstep, std::vector<std::string> ch_names ) :
  frames_name(filename), channel_names(ch_names), frame(new cv::Mat()),
  first_frame(firstf), last_frame(lastf+1), frame_step(fstep)
{
  // set first_frame, frame_step and last_frame from files
  DiscoverFrameSeqNumbering(frames_name, first_frame, frame_step, last_frame);
  Initialize();
}


std::vector<std::string> GetChannelNames( pb::Stream &s )
{
  std::vector<std::string> chnames(s.ExtensionSize(fsio::exr_channel));
  for (size_t ch = 0; ch < chnames.size(); ++ch)
    chnames[ch] = s.GetExtension(fsio::exr_channel, ch);
  return chnames;
}


EXRTextureLoader::EXRTextureLoader( pb::Stream &s, std::string work_dir) :
  frames_name(work_dir + s.video()), channel_names(GetChannelNames(s)),
  frame(new cv::Mat()),
  first_frame(s.HasExtension(fsio::first_frame) ? s.GetExtension(fsio::first_frame) : -1), 
  last_frame (s.HasExtension(fsio:: last_frame) ? s.GetExtension(fsio:: last_frame) +1 :  0),
  frame_step (s.HasExtension(fsio:: frame_step) ? s.GetExtension(fsio:: frame_step) : -1)
{
  // set first_frame, frame_step and last_frame from files
  DiscoverFrameSeqNumbering(frames_name, first_frame, frame_step, last_frame);
  Initialize();

  stream = &s;

  // update protobuf stream
  stream->SetExtension(fsio::first_frame, first_frame);
  stream->SetExtension(fsio::frame_step, frame_step);
  stream->SetExtension(fsio::last_frame, last_frame-1);

  /* if channel_names is empty the full EXR image will loaded
   * in this case, we read the first frame, and add the channel names
   * to the list and the protobuf description of the stream */
  if (stream->ExtensionSize(fsio::exr_channel) == 0)
    for (size_t ch = 0; ch < channel_names.size(); ++ch)
      stream->AddExtension(fsio::exr_channel, channel_names[ch]);
}


EXRTextureLoader::~EXRTextureLoader()
{
  delete frame;
}

