
#include "exr_utils.h"

#include <assert.h>
#include <stdexcept>
// #include <exception>
using namespace std;

// OpenEXR
#include <ImfInputFile.h>
#include <ImfOutputFile.h>
#include <ImfArray.h>
#include <ImfChannelList.h>
#include <Iex.h>		// exceptions
#include <half.h>		// half type
using namespace Imf;
using namespace Imath;

// OpenCV
using namespace cv;


namespace exr_utils
{

// Read EXR file with multiple channels based on the HHI format:
// - 8 channels of type HALF.
// - Same size for all channels.
// - Channel order is: A, B, G, R, Z, Zamplitude, Zflags, Zintensity.
// Returns a list of OpenCV matrices containing the data.
// (Based on http://www.openexr.com/ReadingAndWritingImageFiles.pdf (sec. "Reading an Image File"))
void readEXR_Multichannel_HHI(const string & filename, vector<Mat> & channelData)
{
	// Open EXR file (throws exception if not found)
	InputFile file(filename.c_str());

	// Image size
	Box2i dw = file.header().dataWindow();
	int width = dw.max.x - dw.min.x + 1;
	int height = dw.max.y - dw.min.y + 1;
		
	// Allocate memory buffers to store channel data
	Array2D<half> Ch[8];
	
	// Define framebuffer
	FrameBuffer frameBuffer;

	// Get channel list
	const ChannelList &channels = file.header().channels();

	// Assign buffer for each channel
	unsigned c = 0;
	for (ChannelList::ConstIterator i = channels.begin(); i != channels.end(); ++i, ++c)
	{
		const Channel &channel = i.channel();

// 		// Display channel name and type
// 		cout << i.name() << ", " << (channel.type == 0 ? "uint" : (channel.type == 1 ? "half" : "float")) << endl;
		
		// Check number of channels and data type
		assert(c < 8 && channel.type == HALF);
		
		// Allocate memory buffer
		Ch[c].resizeErase(height, width);
		
		// Insert channel into the framebuffer with its corresponding memory buffer
		frameBuffer.insert(i.name(),
							Slice(	channel.type,					// type
									(char *)(&Ch[c][0][0] - 		// base
											dw.min.x - 
											dw.min.y * width),
									sizeof(Ch[c][0][0]) * 1,		// xStride 
									sizeof(Ch[c][0][0]) * width));	// yStride
	}

	// Read data from file to buffer
	file.setFrameBuffer(frameBuffer);
	file.readPixels(dw.min.y, dw.max.y);
	
	
	// Now convert to OpenCV Matrices
	channelData.resize(8);
	
	for (c = 0; c < 8; ++c)
	{
		channelData[c].create(height, width, CV_32FC1);
		
		for (int y = 0; y < height; ++y)
			for (int x = 0; x < width; ++x)
				channelData[c].at<float>(y,x) = (float)Ch[c][y][x];		// TODO: faster data access
	}
}


// Read EXR file with multiple channels.
// Returns a list of OpenCV matrices containing the data along with a list of channel names.
// All channels are assumed to have the same dimensions.
// Matrix types will be CV_16U for UINT and CV_32F for HALF and FLOAT.
void readEXR_Multichannel(const string & filename, vector<Mat> & channelData, vector<string> & channelNames)
{
	// Open EXR file (throws exception if not found)
	InputFile file(filename.c_str());

	// Image size (same for all channels)
	Box2i dw = file.header().dataWindow();
	int width = dw.max.x - dw.min.x + 1;
	int height = dw.max.y - dw.min.y + 1;
		
	// Memory buffers to store channel data and corresponding types
	vector<char*> memBuff;
	vector<PixelType> types;

	// Define framebuffer
	FrameBuffer frameBuffer;

	// Get channel list from input file
	const ChannelList &channels = file.header().channels();

	// Reset list of channel names
	channelNames.clear();
	
	// For each channel in the list ...
	for (ChannelList::ConstIterator i = channels.begin(); i != channels.end(); ++i)
	{
		const Channel &channel = i.channel();

		// Depth of required memory buffer
		int bytes = (channel.type == HALF ? 2 : 4);
		
		// Allocate memory buffer
		char* buff = new char[height * width * bytes];
		memBuff.push_back(buff);

		// Insert channel into the framebuffer with its corresponding memory buffer
		frameBuffer.insert(i.name(), Slice(channel.type, buff - (dw.min.x + dw.min.y * width) * bytes, bytes, bytes * width));
		
		// Store type to be used when creating the matrices
		types.push_back(channel.type);
		
		// Store channel name to be returned
		channelNames.push_back(i.name());
	}

/*	int ch = 0;
	for (ChannelList::ConstIterator i = channels.begin(); i != channels.end(); ++i, ++ch)
		std::cout << "\t\t\t\tfound channel: " << channelNames[ch] << " of type " << ((i.channel().type == HALF) ? "HALF" : "UINT")  << "\n"; */

	// Read data from file to buffers
	file.setFrameBuffer(frameBuffer);
	file.readPixels(dw.min.y, dw.max.y);

	
	// Now convert to OpenCV Matrices
	channelData.resize(channelNames.size());
	
	// For each channel ...
	vector<PixelType>::const_iterator chtype = types.begin();
	vector<char*>::const_iterator buff = memBuff.begin();
	for (vector<Mat>::iterator chdata = channelData.begin(); chdata !=channelData.end(); ++chdata, ++chtype, ++buff)
	{
		// Create matrix with the corresponding size and type
		int dtype = (*chtype == UINT ? CV_16UC1 : CV_32FC1);
		chdata->create(height, width, dtype);
		
		// Transfer data into the matrix				// TODO: faster data access
		int bytes = (*chtype == HALF ? 2 : 4);
		for (int y = 0, i = 0; y < height; ++y)
			for (int x = 0; x < width; ++x, i += bytes)
				if (*chtype == UINT) chdata->at<unsigned short>(y,x) = (unsigned short)*(unsigned int*)(&(*buff)[i]);
				else if (*chtype == HALF) chdata->at<float>(y,x) = (float)*(half*)(&(*buff)[i]);
				else chdata->at<float>(y,x) = *(float*)(&(*buff)[i]);
	}

	// Deallocate memory buffers
	for (vector<char*>::const_iterator buff = memBuff.begin(); buff != memBuff.end(); ++buff)
		delete [] *buff;
}


// Read EXR file with multiple channels using a list of selected channels.
// Returns a list of OpenCV matrices containing the data.
// All channels are assumed to have the same dimensions.
// Matrix types will be CV_16U for UINT and CV_32F for HALF and FLOAT.
void readEXR_Multichannel(const string & filename, const vector<string> & channelNames, vector<Mat> & channelData)
{
	// Get number of channels from the list
	int nchannels = channelNames.size();
	
	// Open EXR file (throws exception if not found)
	InputFile file(filename.c_str());

	// Image size (same for all channels)
	Box2i dw = file.header().dataWindow();
	int width = dw.max.x - dw.min.x + 1;
	int height = dw.max.y - dw.min.y + 1;
		
	// Memory buffers to store channel data and corresponding types
	vector<char*> memBuff;
	vector<PixelType> types;

	// Define framebuffer
	FrameBuffer frameBuffer;

	// Get channel list from input file
	const ChannelList &channels = file.header().channels();

	// For each channel in the selected list ...
	for (vector<string>::const_iterator chname = channelNames.begin(); chname != channelNames.end(); ++chname)
	{
		// Find channel by name (throws exception if not found)
		const Channel &channel = channels[chname->c_str()];

		// Depth of required memory buffer
		int bytes = (channel.type == HALF ? 2 : 4);
		
		// Allocate memory buffer
		char* buff = new char[height * width * bytes];
		memBuff.push_back(buff);

		// Insert channel into the framebuffer with its corresponding memory buffer
		frameBuffer.insert(chname->c_str(), Slice(channel.type, buff - (dw.min.x + dw.min.y * width) * bytes, bytes, bytes * width));
		
		// Store type to be used when creating the matrices
		types.push_back(channel.type);
	}

	// Read data from file to buffers
	file.setFrameBuffer(frameBuffer);
	file.readPixels(dw.min.y, dw.max.y);

	
	// Now convert to OpenCV Matrices
	channelData.resize(nchannels);
	
	// For each channel ...
	vector<PixelType>::const_iterator chtype = types.begin();
	vector<char*>::const_iterator buff = memBuff.begin();
	for (vector<Mat>::iterator chdata = channelData.begin(); chdata !=channelData.end(); ++chdata, ++chtype, ++buff)
	{
		// Create matrix with the corresponding size and type
		int dtype = (*chtype == UINT ? CV_16UC1 : CV_32FC1);
		chdata->create(height, width, dtype);
		
		// Transfer data into the matrix				// TODO: faster data access
		int bytes = (*chtype == HALF ? 2 : 4);
		for (int y = 0, i = 0; y < height; ++y)
			for (int x = 0; x < width; ++x, i += bytes)
				if (*chtype == UINT) chdata->at<unsigned short>(y,x) = (unsigned short)*(unsigned int*)(&(*buff)[i]);
				else if (*chtype == HALF) chdata->at<float>(y,x) = (float)*(half*)(&(*buff)[i]);
				else chdata->at<float>(y,x) = *(float*)(&(*buff)[i]);
	}

	// Deallocate memory buffers
	for (vector<char*>::const_iterator buff = memBuff.begin(); buff != memBuff.end(); ++buff)
		delete [] *buff;
}


// Read EXR file with multiple channels.
// Returns data as channels in an OpenCV matrix along with a list of channel names,
// the channels in the OpenCV matrix and the names in the list are in the same order.
// All channels are assumed to have the same dimensions and type.
// Matrix type is going to be CV_16U for UINT and CV_32F for HALF and FLOAT.
bool readEXR_MultichannelInterleaved(const std::string& filename,
		cv::Mat& channelData, std::vector<std::string>& ch_names)
{
	std::vector<cv::Mat> channels(0);

	// TODO bool return value for readEXR
	readEXR_Multichannel(filename, channels, ch_names);

	int depth = channels[0].depth();
	int rows  = channels[0].rows;
	int cols  = channels[0].cols;
	for (size_t ch = 0; ch < channels.size(); ++ch)
	{
		if(channels[ch].depth() != depth) 
			throw std::runtime_error("exr_utils::readEXR_MultichannelInterleaved: "
					"specified channels have different pixel depths");
		if(channels[ch].rows != rows || channels[ch].cols != cols)
			throw std::runtime_error("exr_utils::readEXR_MultichannelInterleaved: "
					"specified channels have different sizes");
	}

	// interleave channels into a single Mat
	channelData.create(rows, cols, CV_MAKETYPE(depth, channels.size()));
	cv::merge(channels, channelData);
	
	return true;
}


// Read EXR file with multiple channels using a list of selected channels.
// Returns data as channels in an OpenCV matrix, using the order of the list of channel names.
// All channels are assumed to have the same dimensions and type.
// Matrix type is going to be CV_16U for UINT and CV_32F for HALF and FLOAT.
bool readEXR_MultichannelInterleaved(const std::string& filename,
		const std::vector<std::string>& ch_names, cv::Mat& channelData)
{
	std::vector<cv::Mat> channels(0);

	// TODO bool return value for readEXR
	readEXR_Multichannel(filename, ch_names, channels);

	// failure: some channels were not found
	if (channels.size() != ch_names.size()) return false;

	int depth = channels[0].depth();
	int rows  = channels[0].rows;
	int cols  = channels[0].cols;
	for (size_t ch = 0; ch < channels.size(); ++ch)
	{
		if(channels[ch].depth() != depth) 
			throw std::runtime_error("exr_utils::readEXR_MultichannelInterleaved: "
					"specified channels have different pixel depths");
		if(channels[ch].rows != rows || channels[ch].cols != cols)
			throw std::runtime_error("exr_utils::readEXR_MultichannelInterleaved: "
					"specified channels have different sizes");
	}

	// interleave channels into a single Mat
	channelData.create(rows, cols, CV_MAKETYPE(depth, channels.size()));
	cv::merge(channels, channelData);
	
	return true;
}



// Write EXR file with multiple channels based on the HHI format:
// - 8 channels of type HALF.
// - Same size for all channels.
// - Channel order is: A, B, G, R, Z, Zamplitude, Zflags, Zintensity.
// Channel data given as a list of OpenCV matrices.
// (Based on http://www.openexr.com/ReadingAndWritingImageFiles.pdf (sec. "Writing an Image File"))
void writeEXR_Multichannel_HHI(const string & filename, const vector<Mat> & channelData, Compression compressMethod)
{
	// Check input channels and format
	assert(channelData.size() >= 8 && channelData.front().type() == CV_32FC1);

	// Create header (using first channel dimensions)
	int width = channelData.front().cols;
	int height = channelData.front().rows;
	Header header(width, height);

	// Set compression
	header.compression() = compressMethod;
	
	// Define channels
	header.channels().insert("A", Channel(HALF));
	header.channels().insert("B", Channel(HALF));
	header.channels().insert("G", Channel(HALF));
	header.channels().insert("R", Channel(HALF));
	header.channels().insert("depth.Z", Channel(HALF));
	header.channels().insert("tof.amplitude", Channel(HALF));
	header.channels().insert("tof.flags", Channel(HALF));
	header.channels().insert("tof.intensity", Channel(HALF));

	// Open EXR file
	OutputFile file(filename.c_str(), header);

	// Allocate memory buffers to store channel data
	Array2D<half> Ch[8];

	// Define framebuffer
	FrameBuffer frameBuffer;

	// Get channel list defined above
	const ChannelList &channels = file.header().channels();

	// Assign buffer to each channel
	unsigned c = 0;
	for (ChannelList::ConstIterator i = channels.begin(); i != channels.end(); ++i, ++c)
	{
		const Channel &channel = i.channel();
		
		// Allocate memory buffer
		Ch[c].resizeErase(height, width);

		// Convert OpenCV Matrices to buffers
		for (int y = 0; y < height; ++y)
			for (int x = 0; x < width; ++x)
				Ch[c][y][x] = channelData[c].at<float>(y,x);		// TODO: faster data access

		// Insert channel into the framebuffer with its corresponding memory buffer
		frameBuffer.insert(i.name(), Slice(channel.type, (char *)(&Ch[c][0][0]), sizeof(HALF), sizeof(HALF) * width));
	}

	// Write data from memory to file
	file.setFrameBuffer(frameBuffer);
	file.writePixels(height);
}


// Write EXR file with multiple channels.
// Uses a list of selected channels to be written, consisting of channel names and types.
// Channel data given as a list of OpenCV matrices.
// All channels are assumed to have the same dimensions.
// Matrix types should be CV_16U for UINT and CV_32F for HALF and FLOAT.
void writeEXR_Multichannel(const string & filename, const vector<pair<string, PixelType> > & channelProps, 
						    const vector<Mat> & channelData, Compression compressMethod)
{
	// Get number of channels and resolution from the input list
//	int nchannels = channelData.size();
	int width = channelData.front().cols;		// TODO: support channel-dependent sizes?
	int height = channelData.front().rows;
	
	// Create header with basic information
	Header header(width, height);
	header.compression() = compressMethod;

	// Memory buffers to store channel data
	vector<char*> memBuff;

	// Define framebuffer
	FrameBuffer frameBuffer;

	// For each channel inside the list ...
	vector<pair<string, PixelType> >::const_iterator chprop = channelProps.begin();
	vector<Mat>::const_iterator chdata = channelData.begin();
	for (; chprop != channelProps.end(); ++chprop, ++chdata)
	{
		// Name, type and depth of required memory buffer
		string name = chprop->first;
		PixelType type = chprop->second;
		int bytes = (type == HALF ? 2 : 4);
		
		// Check number of channels and data type for the current channel image
		int dtype = chdata->type();
		assert(chdata->channels() == 1 && ((dtype == CV_16U && type == UINT) || (dtype == CV_32F && type != UINT)));

		// Insert channel into header
		header.channels().insert(name.c_str(), Channel(type));

		// Allocate memory buffer
		char* buff = new char[height * width * bytes];
		memBuff.push_back(buff);

		// Transfer data into the buffer				// TODO: faster data access
		for (int y = 0, i = 0; y < height; ++y)
			for (int x = 0; x < width; ++x, i += bytes)
				if (type == UINT) *(unsigned int*)(&buff[i]) = (unsigned int)chdata->at<unsigned short>(y,x);	
				else if (type == HALF) *(half*)(&buff[i]) = (half)chdata->at<float>(y,x);
				else *(float*)(&buff[i]) = chdata->at<float>(y,x);

		// Insert channel into the framebuffer with its corresponding memory buffer
		frameBuffer.insert(name.c_str(), Slice(type, buff, bytes, bytes * width));
	}

	// Open EXR file
	OutputFile file(filename.c_str(), header);

	// Write data from memory to file
	file.setFrameBuffer(frameBuffer);
	file.writePixels(height);
	
	// Deallocate memory buffers
	for (vector<char*>::const_iterator buff = memBuff.begin(); buff != memBuff.end(); ++buff)
		delete [] *buff;
}


// Write EXR file with multiple channels without truncating the file.
// Uses a list of selected channels to be written, consisting of channel names and types.
// Channel data given as a list of OpenCV matrices.
// All channels are assumed to have the same dimensions.
// Matrix types should be CV_16U for UINT and CV_32F for HALF and FLOAT.
void writeEXR_Multichannel_Append(const string & filename, const vector<pair<string, PixelType> > & channelProps, 
								   const vector<Mat> & channelData, Compression compressMethod)
{
	try
	{
		//// First open file and read existing channels
		
		// Open EXR file (throws exception if not found)
		InputFile file(filename.c_str());
		
		// Image size (same for all channels)
		Box2i dw = file.header().dataWindow();
		int width = dw.max.x - dw.min.x + 1;
		int height = dw.max.y - dw.min.y + 1;
			
		// Memory buffers to store channel data
		vector<char*> memBuff;
		
		// Channels names and types
		vector<string> names;
		vector<PixelType> types;

		// Define framebuffer
		FrameBuffer frameBuffer;

		// Get channel list from input file
		const ChannelList &channels = file.header().channels();

		// For each channel in the list ...
		for (ChannelList::ConstIterator i = channels.begin(); i != channels.end(); ++i)
		{
			const Channel &channel = i.channel();

			// Depth of required memory buffer
			int bytes = (channel.type == HALF ? 2 : 4);
			
			// Allocate memory buffer
			char* buff = new char[height * width * bytes];
			memBuff.push_back(buff);

			// Insert channel into the framebuffer with its corresponding memory buffer
			frameBuffer.insert(i.name(), Slice(channel.type, buff - (dw.min.x + dw.min.y * width) * bytes, bytes, bytes * width));
			
			// Store name and type for its later use
			names.push_back(i.name());
			types.push_back(channel.type);
		}

		// Read data from file to buffers
		file.setFrameBuffer(frameBuffer);
		file.readPixels(dw.min.y, dw.max.y);
		
		
		
		///// Now add/replace selected channels
		
		// Check channel resolution
		assert(width == channelData.front().cols && height == channelData.front().rows);
		
		// Create header
		Header header(width, height);
		header.channels() = file.header().channels();	// copy channels
		header.compression() = compressMethod;			// replace compression

		// For each channel inside the list ...
		vector<pair<string, PixelType> >::const_iterator chprop = channelProps.begin();
		vector<Mat>::const_iterator chdata = channelData.begin();
		for (; chprop != channelProps.end(); ++chprop, ++chdata)
		{
			// Name, type and depth of required memory buffer
			string name = chprop->first;
			PixelType type = chprop->second;
			int bytes = (type == HALF ? 2 : 4);
			
			// Check number of channels and data type for the current channel image
			int dtype = chdata->type();
			assert(chdata->channels() == 1 && ((dtype == CV_16U && type == UINT) || (dtype == CV_32F && type != UINT)));

			char* buff;
			
			// Add channel if not present
			Channel* ch = header.channels().findChannel(name.c_str());
			if (!ch)
			{
				// Insert channel into header
				header.channels().insert(name.c_str(), Channel(type));

				// Allocate memory buffer
				buff = new char[height * width * bytes];
				memBuff.push_back(buff);
				
				// Insert channel into the framebuffer with its corresponding memory buffer
				frameBuffer.insert(name.c_str(), Slice(type, buff, bytes, bytes * width));
			}
			else
			{
				// Find memory buffer from channel name
				int c = find(names.begin(), names.end(), name) - names.begin();
				buff = memBuff[c];
				
				// Reallocate things if types differ
				if (ch->type != type)
				{
					// Replace channel type
					ch->type = type;

					// Reallocate memory buffer
					delete [] buff;
					buff = new char[height * width * bytes];
					
					// Replace channel/slice info in the framebuffer
					frameBuffer[name.c_str()] = Slice(type, buff, bytes, bytes * width);
				}
			}

			// Transfer data into the buffer				// TODO: faster data access
			for (int y = 0, i = 0; y < height; ++y)
				for (int x = 0; x < width; ++x, i += bytes)
					if (type == UINT) *(unsigned int*)(&buff[i]) = (unsigned int)chdata->at<unsigned short>(y,x);	
					else if (type == HALF) *(half*)(&buff[i]) = (half)chdata->at<float>(y,x);
					else *(float*)(&buff[i]) = chdata->at<float>(y,x);
		}

		// Open EXR file for writting
		OutputFile ofile(filename.c_str(), header);

		// Write data from memory to file
		ofile.setFrameBuffer(frameBuffer);
		ofile.writePixels(height);
		
		// Deallocate memory buffers
		for (vector<char*>::const_iterator buff = memBuff.begin(); buff != memBuff.end(); ++buff)
			delete [] *buff;
	}
	// If an exception is thrown because of input file not existing ...
// 	catch (const exception &exc)
	catch (const Iex::ErrnoExc &exc)
	{
		// Just create a new file
		writeEXR_Multichannel(filename, channelProps, channelData, compressMethod);

// 		cout << exc.what() << endl;
	}
}


// Write EXR file with multiple channels.
// Uses a list of selected channels to be written.
// Channel data given as channels of an OpenCV matrix.
// EXR output type is either UINT for CV_16U input type, or HALF for CV_32F type.
void writeEXR_MultichannelInterleaved(const std::string & filename,
		const std::vector< std::string > & channelNames, const cv::Mat& channelData,
		Imf::Compression compressMethod)
{
	if((size_t)channelData.channels() != channelNames.size())
		throw std::runtime_error("exr_utils::writeEXR_MultichannelInterleaved: "
				"number of channels does not match number of channel names");

	std::vector<cv::Mat> channels(channelData.channels());
	cv::split(channelData,channels);

	// build list of pairs
	std::vector< std::pair<std::string, Imf::PixelType> > channelProps(channelNames.size());
	Imf::PixelType type;
	switch (channelData.depth())
	{
		case CV_16U: type = UINT; break;
		case CV_32F: type = HALF; break;
		default: 
			throw std::runtime_error("exr_utils::writeEXR_MultichannelInterleaved: "
				"input channel with unsupported type.");
	}

	for (size_t ch = 0; ch < channelNames.size(); ++ch)
		channelProps[ch] = std::pair<std::string, Imf::PixelType>(channelNames[ch], type);

	writeEXR_Multichannel(filename, channelProps, channels, compressMethod);
}



// Append channels into an existing multichannel EXR file.
// Uses a list of selected channels to be written.
// Channel data given as channels of an OpenCV matrix.
// EXR output type is either UINT for CV_16U input type, or HALF for CV_32F type.
void appendEXR_MultichannelInterleaved(const std::string & filename,
		const std::vector< std::string > & ch_names, const cv::Mat& ch_data,
		Imf::Compression compressMethod)
#if 1
{
	if((size_t)ch_data.channels() != ch_names.size())
		throw std::runtime_error("exr_utils::appendEXR_MultichannelInterleaved: "
				"number of channels does not match number of channel names");

	std::vector<cv::Mat> channels(ch_data.channels());
	cv::split(ch_data,channels);

	// build list of pairs
	std::vector< std::pair<std::string, Imf::PixelType> > ch_props(ch_names.size());
	{
		Imf::PixelType type;
		switch (ch_data.depth())
		{
			case CV_16U: type = UINT; /*std::cout << "\t\t\t\tadd CV_16U\n";*/ break;
			case CV_32F: type = HALF; /*std::cout << "\t\t\t\tadd CV_32F\n";*/ break;
			default:
				throw std::runtime_error("exr_utils::writeEXR_MultichannelInterleaved: "
					"input channel with unsupported type.");
		}

		for (size_t ch = 0; ch < ch_names.size(); ++ch)
			ch_props[ch] = std::pair<std::string, Imf::PixelType>(ch_names[ch], type);
	}

	// try to read file
	std::vector<cv::Mat> found_channels(0);
	std::vector<std::string> found_ch_names(0);
	try { readEXR_Multichannel(filename, found_channels, found_ch_names); }
	catch (...) { }

	// build list of pairs for found channels
	std::vector< std::pair<std::string, Imf::PixelType> > found_ch_props(found_ch_names.size());
	{
		for (size_t ch = 0; ch < found_ch_names.size(); ++ch)
		{
			Imf::PixelType found_type;
			switch (found_channels[ch].depth())
			{
				case CV_16U: found_type = UINT; /*std::cout << "\t\t\t\t" + found_ch_names[ch] + " found CV_16U\n";*/ break;
				case CV_32F: found_type = HALF; /*std::cout << "\t\t\t\t" + found_ch_names[ch] + " found CV_32F\n";*/ break;
				default: throw std::runtime_error("exr_utils::writeEXR_MultichannelInterleaved: "
							 "read existing channel with unsupported type.");
			}
			found_ch_props[ch] = std::pair<std::string, Imf::PixelType>(found_ch_names[ch], found_type);
		}
	}
	

	// add found_ch_props and found_channels to ch_props and channels
	{
		for (size_t ch = 0; ch < found_ch_props.size(); ++ch)
			if (std::find(ch_names.begin(), ch_names.end(), found_ch_names[ch]) == ch_names.end() )
			{
				ch_props.push_back(found_ch_props[ch]);
				channels.push_back(found_channels[ch]);
			}
	}
	
	// write exr image 
	writeEXR_Multichannel(filename, ch_props, channels, compressMethod);
}
#else
{
	// TODO throw exception
	assert((size_t)channelData.channels() == channelNames.size());

	std::vector<cv::Mat> channels(channelData.channels());
	cv::split(channelData,channels);

	// build list of pairs
	std::vector< std::pair<std::string, Imf::PixelType> > channelProps(channelNames.size());
	Imf::PixelType type;
	switch (channelData.depth())
	{
		case CV_16U: type = UINT; break;
		case CV_32F: type = HALF; break;
		default: break;// TODO throw invalid type exception
	}

	for (size_t ch = 0; ch < channelNames.size(); ++ch)
		channelProps[ch] = std::pair<std::string, Imf::PixelType>(channelNames[ch], type);

	writeEXR_Multichannel_Append(filename, channelProps, channels);
}
#endif




}
