#ifndef EXR_UTILS_H
#define EXR_UTILS_H

#include <string>
#include <vector>
#include <utility>				// pair
#include <opencv2/opencv.hpp>	// OpenCV's Mat
#include <ImfInputFile.h>		// OpenEXR's flags


namespace exr_utils
{
	// Read EXR file with multiple channels based on the HHI format:
	// - 8 channels of type HALF.
	// - Same size for all channels.
	// - Channel order is: A, B, G, R, Z, Zamplitude, Zflags, Zintensity.
	// Returns a list of OpenCV matrices containing the data.
	// (Based on http://www.openexr.com/ReadingAndWritingImageFiles.pdf (sec. "Reading an Image File"))
	void readEXR_Multichannel_HHI(const std::string & filename, std::vector<cv::Mat> & channelData);
	
	// Read EXR file with multiple channels.
	// Returns a list of OpenCV matrices containing the data along with a list of channel names.
	// All channels are assumed to have the same dimensions.
	// Matrix types will be CV_16U for UINT and CV_32F for HALF and FLOAT.
	void readEXR_Multichannel(const std::string & filename, std::vector<cv::Mat> & channelData, std::vector<std::string> & channelNames);

	// Read EXR file with multiple channels using a list of selected channels.
	// Returns a list of OpenCV matrices containing the data.
	// All channels are assumed to have the same dimensions.
	// Matrix types will be CV_16U for UINT and CV_32F for HALF and FLOAT.
	void readEXR_Multichannel(const std::string & filename, const std::vector<std::string> & channelNames, std::vector<cv::Mat> & channelData);

	// Read EXR file with multiple channels.
	// Returns data as channels in an OpenCV matrix along with a list of channel names,
	// the channels in the OpenCV matrix and the names in the list are in the same order.
	// All channels are assumed to have the same dimensions and type.
	// Matrix type is going to be CV_16U for UINT and CV_32F for HALF and FLOAT.
	bool readEXR_MultichannelInterleaved(const std::string& filename, cv::Mat& channelData, std::vector<std::string>& channelNames);

	// Read EXR file with multiple channels using a list of selected channels.
	// Returns data as channels in an OpenCV matrix, using the order of the list of channel names.
	// All channels are assumed to have the same dimensions and type.
	// Matrix type is going to be CV_16U for UINT and CV_32F for HALF and FLOAT.
	bool readEXR_MultichannelInterleaved(const std::string& filename, const std::vector<std::string>& channelNames, cv::Mat& channelData);
	
	
	// Write EXR file with multiple channels based on the HHI format:
	// - 8 channels of type HALF.
	// - Same size for all channels.
	// - Channel order is: A, B, G, R, Z, Zamplitude, Zflags, Zintensity.
	// Channel data given as a list of OpenCV matrices.
	// (Based on http://www.openexr.com/ReadingAndWritingImageFiles.pdf (sec. "Writing an Image File"))
	void writeEXR_Multichannel_HHI(const std::string & filename, const std::vector<cv::Mat> & channelData, Imf::Compression compressMethod = Imf::PIZ_COMPRESSION);
	
	// Write EXR file with multiple channels.
	// Uses a list of selected channels to be written, consisting of channel names and types.
	// Channel data given as a list of OpenCV matrices.
	// All channels are assumed to have the same dimensions.
	// Matrix types should be CV_16U for UINT and CV_32F for HALF and FLOAT.
	void writeEXR_Multichannel(const std::string & filename, const std::vector< std::pair<std::string,Imf::PixelType> > & channelProps, 
							    const std::vector<cv::Mat> & channelData, Imf::Compression compressMethod = Imf::PIZ_COMPRESSION);
	
	// Write EXR file with multiple channels without truncating the file.
	// Uses a list of selected channels to be written, consisting of channel names and types.
	// Channel data given as a list of OpenCV matrices.
	// All channels are assumed to have the same dimensions.
	// Matrix types should be CV_16U for UINT and CV_32F for HALF and FLOAT.
	void writeEXR_Multichannel_Append(const std::string & filename, const std::vector< std::pair<std::string,Imf::PixelType> > & channelProps, 
									   const std::vector<cv::Mat> & channelData, Imf::Compression compressMethod = Imf::PIZ_COMPRESSION);

	// Write EXR file with multiple channels.
	// Uses a list of selected channels to be written.
	// Channel data given as channels of an OpenCV matrix.
	// EXR output type is either UINT for CV_16U input type, or HALF for CV_32F type.
	void writeEXR_MultichannelInterleaved(const std::string & filename, const std::vector< std::string > & channelNames, 
							    const cv::Mat& channelData, Imf::Compression compressMethod = Imf::PIZ_COMPRESSION);
	
	// Append channels into an existing multichannel EXR file.
	// Uses a list of selected channels to be written.
	// Channel data given as channels of an OpenCV matrix.
	// EXR output type is either UINT for CV_16U input type, or HALF for CV_32F type.
	void appendEXR_MultichannelInterleaved(const std::string & filename, const std::vector< std::string > & channelNames, 
									   const cv::Mat& channelData, Imf::Compression compressMethod = Imf::PIZ_COMPRESSION);

}

#endif	// EXR_UTILS_H
