#include "frames_texture_loader.h"
#include <stdexcept>
#include <iostream>
#include <fstream>
#include <cmath>

#include <scene-persistence/autogen/scenedb.pb.h>
#include <scene-persistence/autogen/stream_fsio.pb.h>

// opencv is only used to read the frames
#include <opencv2/opencv.hpp>

#include "frame_seq_utils.h"


bool FramesTextureLoader::NextFrame()
{
	current_frame += frame_step;
 
	if (current_frame < last_frame)
	{
		// form frame's filename
		char * name = new char[frames_name.size() + 10];
		std::sprintf(name,frames_name.c_str(),current_frame);
		
		// open image and set video_buffer
		*frame = cv::imread( name, CV_LOAD_IMAGE_UNCHANGED );
		delete [] name;

		video_buffer = (uint8_t *)frame->data;


		if (!frame->data) return false;

		if (frame->cols != width || frame->rows != height || frame->channels() != channels)
			throw std::runtime_error("Loaded frame size/channels do not match FramesTextureLoader's");
		if ((int)frame->elemSize1() != depth)
			throw std::runtime_error("Loaded frame pixel depth do not match FramesTextureLoader's");

		return true;
	}

	// current_frame >= last_frame
	return false;
}


void FramesTextureLoader::SetCurrentData(const uint8_t *data, int w, int h, int c, int d)
{
	if (current_frame < first_frame)
		throw std::runtime_error("current_frame < first_frame in FramesTextureLoader::SetCurrentData");

	// for a newly created file stream, dimensions are not initialized yet
	if (!(width > 0 && height > 0 && channels > 0 && depth > 0))
	{
		if (!(w > 0 && h > 0 && c > 0 && d > 0))
			throw std::runtime_error("Calling FramesTextureLoader::SetCurrentData "
					"without dimensions/channels/pixel-depth for a FramesTextureLoader "
					"whose dimensions/channel/pixel-depth have not yet been initialized");

		width = w;
		height = h;
		channels = c;
		depth = d;
	}
	// if dimensions have already been initialized, input dimensions should match them
	else
	{
		if (w > 0 && w != width   ) throw std::runtime_error("Input width doesn't match FramesTextureLoader's width");
		if (h > 0 && h != height  ) throw std::runtime_error("Input height doesn't match FramesTextureLoader's height");
		if (c > 0 && c != channels) throw std::runtime_error("Input channels don't match FramesTextureLoader's channels");
		if (d > 0 && d != depth   ) throw std::runtime_error("Input pixel-depth doesn't match FramesTextureLoader's pixel-depth");
	}



	// form frame's filename
	char * name = new char[frames_name.size() + 10];
	std::sprintf(name,frames_name.c_str(),current_frame);

	// write image
	int type;
	switch (depth)
	{
		case 1: type = CV_MAKETYPE(CV_8U,channels); break;
		case 2: type = CV_MAKETYPE(CV_16U,channels); break;
		default: throw std::runtime_error("Unsupported pixel-depth in FramesTextureLoader");
	}
	cv::imwrite( name, cv::Mat(height, width, type, (void *)data) );
	delete [] name;

	// if current frame exceeds last_frame, update last_frame
	if (current_frame == last_frame)
	{
		++last_frame;
		if (stream) stream->SetExtension(fsio::last_frame, last_frame-1);
	}
}


void FramesTextureLoader::Initialize()
{
	// tries to set current_frame, width, height, channels and depth by reading
	// the first file in the file sequence.  This function is used by the
	// constructors only. 
 
	// set current frame
	current_frame = first_frame - 1;

	// open first frame to set height, width and channels
	{
		// form frame's filename
		char * name = new char[frames_name.size() + 10];
		std::sprintf(name,frames_name.c_str(),first_frame);
		
		// open image and set video_buffer
		cv::Mat f = cv::imread( name, CV_LOAD_IMAGE_UNCHANGED );
		delete [] name;

		height   = f.data ? f.rows        : -1;
		width    = f.data ? f.cols        : -1;
		channels = f.data ? f.channels()  : -1;
		depth    = f.data ? f.elemSize1() : -1;
	}
}
	


FramesTextureLoader::FramesTextureLoader( std::string filename, 
    int firstf, int lastf, int fstep ) :
  frames_name(filename), frame(new cv::Mat()),
  first_frame(firstf), last_frame(lastf+1), frame_step(fstep)
{
	DiscoverFrameSeqNumbering(frames_name, first_frame, frame_step, last_frame);
	Initialize();
}


FramesTextureLoader::FramesTextureLoader( pb::Stream & s, std::string work_dir) :
  frames_name(work_dir + s.video()),
  frame(new cv::Mat()),
  first_frame(s.HasExtension(fsio::first_frame) ? s.GetExtension(fsio::first_frame) : -1), 
  last_frame (s.HasExtension(fsio:: last_frame) ? s.GetExtension(fsio:: last_frame) +1 :  0),
  frame_step (s.HasExtension(fsio:: frame_step) ? s.GetExtension(fsio:: frame_step) : -1)
{
	DiscoverFrameSeqNumbering(frames_name, first_frame, frame_step, last_frame);
	Initialize();

  stream = &s;

  // update protobuf stream
  stream->SetExtension(fsio::first_frame, first_frame);
  stream->SetExtension(fsio::frame_step, frame_step);
  stream->SetExtension(fsio::last_frame, last_frame-1);
}


FramesTextureLoader::~FramesTextureLoader()
{
	delete frame;
}

