#include "oflow_texture_loader.h"
#include <stdexcept>
#include <iostream>
#include <fstream>
#include <cmath>

#include <scene-persistence/autogen/scenedb.pb.h>
#include <scene-persistence/autogen/stream_fsio.pb.h>

// opencv is only used to read the frames
#include <opencv2/opencv.hpp>

#include "frame_seq_utils.h"

namespace cv_utils
{
	template <typename T>
	int read_ascii_2D(const std::string name, cv::Mat& mat, bool size_header = true)
	{
		FILE *f = fopen(name.c_str(),"rt");

		// error checkings
		{
			if (f == NULL) CV_Error( CV_StsObjectNotFound, "file not found");
			if (mat.dims > 2) CV_Error( CV_StsBadSize, "function only handles 2D matrices" );
			if (mat.channels() > 1) CV_Error ( CV_StsBadSize, "function only handles scalar matrices" );
		}

		int sz[2] = {0,0};
		if (size_header)
		{
			int tmp1 = 0,tmp2 = 0;
			int success = fscanf(f,"%d %d %d %d\n",sz+1,sz,&tmp1,&tmp2);
			if (!success) CV_Error( CV_StsError, "unexpected format when reading header");
			if ((mat.size[0] != 0) &&
				((mat.size[0] != sz[0]) || (mat.size[1] != sz[1])))
				CV_Error( CV_StsBadSize, "expected image size differs from header size" );
		}
		else if (mat.size[0] == 0) // no size header (matlab ascii format) - determine size
		{
			// number of rows
			int c;
			while (EOF != (c = fgetc(f))) if(c =='\n') ++sz[0];

			// number of cols
			rewind(f);
			for (sz[1] = 1; sz[1] < 2000; ++sz[1])
			{
				// we expect a floating point number
				float data;
				int success = fscanf(f,"%g",&data);
				if (!success) CV_Error( CV_StsError, "unexpected format when reading number");

				// read next character
				c = fgetc(f);
				if ((c == '\n') || (c == EOF)) break;
				if (c != ' ') CV_Error( CV_StsError, "unexpected format after number\n");
			}
			rewind(f);
		}
		else
		{
			sz[0] = mat.size[0];
			sz[1] = mat.size[1];
		}


		// allocate a Mat file
		mat.create(sz[0],sz[1],cv::DataType<T>::type);

		T *p_mat = (T *)mat.data;
		for (int i0 = 0; i0 < sz[0]; ++i0)
		{
			for (int i1 = 0; i1 < sz[1]; ++i1, ++p_mat)
			{
				float data;
				int success = fscanf(f,"%f",&data);
				if (!success) CV_Error( CV_StsError, "unexpected format when reading number");
				*p_mat = (T)data;
			}

			int c = fgetc(f);
			if ((c != '\n') && (c == EOF)) CV_Error( CV_StsError, "unexpected format after number\n");
		}

		fclose(f);
		return 1;
	}
}


bool OFlowTextureLoader::NextFrame()
{
	current_frame += frame_step;
 
	if (current_frame < last_frame)
	{
		char * name_root = new char[flows_path.size() + 10];
		std::sprintf(name_root,flows_path.c_str(),current_frame);

		// form _u component's filename
		{
			std::string name(name_root);
			name += "_u.oflow";
			
			// open image and set video_buffer
			cv_utils::read_ascii_2D<float>(name,flow_components[0]);

			assert(flow_components[0].cols == width && 
			       flow_components[0].rows == height);
		}

		// form _v component's filename
		{
			std::string name(name_root);
			name += "_v.oflow";
			
			// open image and set video_buffer
			cv_utils::read_ascii_2D<float>(name,flow_components[1]);

			assert(flow_components[1].cols == width && 
			       flow_components[1].rows == height);
		}

		// copy flow_components as channels in flow
		cv::merge(flow_components,2,*flow);

		video_buffer = (uint8_t *)flow->data;

		delete [] name_root;
		return true;
	}
	
	// current_frame >= last_frame
	return false;
}


void OFlowTextureLoader::Initialize(const std::string filename)
{
	// set current frame
	current_frame = first_frame - 1;

	// open first frame to set height and width
	{
		// form frame's filename
		char * name = new char[filename.size() + 10];
		std::sprintf(name,filename.c_str(),first_frame);
		
		cv_utils::read_ascii_2D<float>(name,flow_components[0]);
		delete [] name;

		height = flow_components[0].rows;
		width  = flow_components[0].cols;
		channels = 2;
	}


	// set flows_path
	{
		size_t i0 = filename.find('%');
		size_t i1 = filename.rfind("_u.oflow");

		if (i1 == std::string::npos) i1 = filename.rfind("_v.oflow",i0);

		// TODO throw exeception when file not found!

		flows_path = filename.substr(0,i1);

		/*
		std::cout << filename << std::endl 
		          << "|filename| = " << filename.length() 
		          << "  i1 = " << i1 << std::endl;
		std::cout << "|flows_path| = " << flows_path.length() << std::endl;
		std::cout << flows_path << std::endl;
		*/
	}
}



OFlowTextureLoader::OFlowTextureLoader( std::string filename, 
    int firstf, int lastf, int fstep ) :
  flow(new cv::Mat()), flow_components(new cv::Mat[2]),
  first_frame(firstf), last_frame(lastf+1), frame_step(fstep)
{
	DiscoverFrameSeqNumbering(filename, first_frame, frame_step, last_frame);
	Initialize(filename);
}




OFlowTextureLoader::OFlowTextureLoader( pb::Stream & s, std::string work_dir) :
  flow(new cv::Mat()), flow_components(new cv::Mat[2]),
  first_frame(s.HasExtension(fsio::first_frame) ? s.GetExtension(fsio::first_frame) : -1), 
  last_frame (s.HasExtension(fsio:: last_frame) ? s.GetExtension(fsio:: last_frame) +1 :  0),
  frame_step (s.HasExtension(fsio:: frame_step) ? s.GetExtension(fsio:: frame_step) : -1)
{
	DiscoverFrameSeqNumbering(work_dir + s.video(), first_frame, frame_step, last_frame);
	Initialize(work_dir + s.video());

	stream = &s;
}


OFlowTextureLoader::~OFlowTextureLoader()
{
	delete flow;
	delete [] flow_components;
}

