#ifndef TEXTURE_LOADER_API_H
#define TEXTURE_LOADER_API_H

#include "stdint.h"
#include <string>
#include <vector>
#include <memory>

namespace pb
{
  class Stream;
}

class ITextureLoader
{
	protected:

		uint8_t *video_buffer;

		/* The texture loaders handle both the `physical` stream
		 * and the protobuf representation. This design breaks
		 * modularity in the code (texture loaders should be
		 * independent of the data serialization library), but 
		 * it is simpler. */
		pb::Stream * stream;

		int current_frame;
		int width;
		int height;
		int channels;
		int depth; // bytes per pixel per channel (eg. 1 for RGB with 8 bits)

	public:


		// ctor with no arguments is called automatically by chidrens' ctors
		ITextureLoader(void) : 
			video_buffer(0), stream(0), current_frame(-1), width(-1), height(-1),
			channels(-1), depth(-1) {};

		int Width(void) const { return width; };
		int Height(void) const { return height; };
		int Channels(void) const { return channels; };
		int PixelDepth(void) const { return depth; };

		// returns frame index
		int FrameIndex(void) const { return current_frame; };

		// read next frame and update position in the stream
		virtual bool NextFrame() = 0;

		// returns width*height array of RGB (unsigned bytes)
		uint8_t* GetCurrentData() { return video_buffer; }

		// writes provided data into stream
		virtual void SetCurrentData(const uint8_t* data, int w = -1, int h = -1, int c = -1, int d = -1) = 0;

		// return size in bytes
		size_t Size() { return width * height * channels * depth; }

		// factory
		static std::shared_ptr<ITextureLoader> CreateTextureLoader(std::string file,
				int first_frame = -1, int last_frame = -1, int frame_step = 1,
				std::vector<std::string> channel_names = std::vector<std::string>(0));

		// factory
		static std::shared_ptr<ITextureLoader> CreateTextureLoader(pb::Stream & s,
				std::string work_dir);

		~ITextureLoader(){}

};


#endif
