/*
 * LoadCameraProjectionTest.cpp
 *
 *  Created on: Jun 04, 2014
 *      Author: Michael Reinhardt
 */

// this repository
#include "scene-core/sra-api/sra_scene.h"
#include "scene-core/sra_implementation.h"
//#include "scene-persistence/backend/IdManager.h"
#include "scene-persistence/backend/sraBackend.h"

// operating system
#include <iostream>
#include <string>
#include <array>

static const bool g_bUseComplexTest = true;
static const std::string g_szSceneFile = "WriteCameraProjectionTest.scene";

// Unit test for loading a camera node from various locations inside a scene graph.

/********************************************************************************************/

bool performTest(const std::string szTestMessage,
								 PScene spcScene,
								 const int iParentId,
								 const std::string &szCameraNameExpected,
								 const sra::float_buffer_t &cProjectionMatrixExpected,
								 const std::vector<std::string> &vszCorrespondingStreamsExpected)
{
	// Define variables for return values
	std::cout << "TEST: " << szTestMessage << std::endl;
	// paranoia checks
	if(true == szCameraNameExpected.empty())
	{
		std::cerr << "Expected camera name is empty. Aborting test." << std::endl;
		return false;
	}
	if(12 != cProjectionMatrixExpected.Size())
	{
		std::cerr << "Expected projection matrix has wrong size of " << cProjectionMatrixExpected.Size() << ". Required is 12. Aborting test." << std::endl;
		return false;
	}
	if(NULL == cProjectionMatrixExpected.Ptr())
	{
		std::cerr << "Expected projection matrix has invalid data pointer: " << cProjectionMatrixExpected.Ptr() << " Aborting test." << std::endl;
		return false;
	}

	try {
		bool bTestResult = true;
		std::string szCameraName;
		sra::float_buffer_t cProjectionMatrix(0, NULL);
		std::vector<std::string> vszCorrespondingStreams;

		if(false == backend::getCameraProjectionMatrix(spcScene, iParentId, szCameraName, cProjectionMatrix, vszCorrespondingStreams))
		{
			std::cerr << "TEST FAILED: getCameraProjectionMatrix returned false" << std::endl;
			bTestResult = false;
		}
		// check return values
		if(szCameraNameExpected.compare(szCameraName) != 0)
		{
			std::cerr << "TEST FAILED: camera name is wrong. Expected: " << szCameraNameExpected << " Found: " << szCameraName << std::endl;
			bTestResult = false;
		}
		if(cProjectionMatrixExpected.Size() != cProjectionMatrix.Size())
		{
			std::cerr << "TEST FAILED: returned projection matrix has wrong size. Expected: " << cProjectionMatrixExpected.Size() <<
				" Found: " << cProjectionMatrix.Size() << std::endl;
			bTestResult = false;
		}
		else
		{
			// This check relies on the correct size of the returned projection matrix.
			if(0 != std::memcmp(cProjectionMatrixExpected.Ptr(), cProjectionMatrix.Ptr(), cProjectionMatrixExpected.Size()))
			{
				std::cerr << "TEST FAILED: returned projection matrix has wrong data." << std::endl;
				std::cerr << "Expected\t\t\tFound" << std::endl;
				for (int index = 0; index < static_cast<int>(cProjectionMatrixExpected.Size()); ++index)
				{
					std::cerr << cProjectionMatrixExpected.Ptr()[index] << "\t\t" << cProjectionMatrix.Ptr()[index] << std::endl;
				}
				bTestResult = false;
			}
		}
		if(vszCorrespondingStreamsExpected != vszCorrespondingStreams)
		{
			std::cerr << "TEST FAILED: returned stream names do not match expected names." << std::endl;
			std::cerr << "Expected      Found" << std::endl;
			for (int index = 0; index < static_cast<int>(vszCorrespondingStreamsExpected.size()); ++index)
			{
				std::cerr << vszCorrespondingStreamsExpected[index] << "  " << vszCorrespondingStreams[index] << std::endl;
			}
			bTestResult = false;
		}

		if(true == bTestResult)
		{
			std::cout << "TEST SUCCEEDED" << std::endl;
		}
		std::cout << "---------------------------------------------." << std::endl;
		return bTestResult;
	}
	catch(const std::runtime_error& e)
	{
		std::cerr << "TEST FAILED: " << e.what() << std::endl;
		std::cout << "---------------------------------------------." << std::endl;
		return false;
	}
}



bool simpleTest(PScene spcScene, PAcel spcGlobals)
{
	// Define variables for expected values
	int iParendId;
	std::string szCameraName;
	const std::array<float, 12> afProjMatrix = {11, 12, 13, 14,
	                                            21, 22, 23, 24,
	                                            31, 32, 33, 34 };
	sra::float_buffer_t cProjectionMatrix(12, afProjMatrix.data());
	std::vector<std::string> vszStreamNames;
	vszStreamNames.push_back("The");
	vszStreamNames.push_back("weather");
	vszStreamNames.push_back("is");
	vszStreamNames.push_back("too good");
	vszStreamNames.push_back("for coding.");

	// Run tests
	iParendId = 3;
	szCameraName = "Zero";
	performTest("Lookup: camera acel at top level.",
		spcScene, iParendId, szCameraName, cProjectionMatrix, vszStreamNames);
	return true;
}

bool complexTest(PScene spcScene, PAcel spcGlobals)
{
	// Currently one the non-property tests are checked. When these will work someday, the Id will change.
	// This will require some refactoring in this function.

	// Define variables for expected values
	int iParendId;
	std::string szCameraName;
	const std::array<float, 12> afProjMatrix = {11, 12, 13, 14,
	                                            21, 22, 23, 24,
	                                            31, 32, 33, 34 };
	sra::float_buffer_t cProjectionMatrix(12, afProjMatrix.data());
	std::vector<std::string> vszStreamNames;
	vszStreamNames.push_back("No");
	vszStreamNames.push_back("Way");

	iParendId = 19;
	szCameraName = "One";
	performTest("Lookup: camera acel of a first level child.",
		spcScene, iParendId, szCameraName, cProjectionMatrix, vszStreamNames);

	iParendId = -1;
	szCameraName = "Two";
	performTest("Lookup: camera acel of a second level property (first level: child).",
		spcScene, iParendId, szCameraName, cProjectionMatrix, vszStreamNames);

	iParendId = 27;
	szCameraName = "Three";
	performTest("Lookup: camera acel of a second level child (first level: child)",
		spcScene, iParendId, szCameraName, cProjectionMatrix, vszStreamNames);

	iParendId = -1;
	szCameraName = "Four";
	performTest("Lookup: camera acel of a first level property.",
		spcScene, iParendId, szCameraName, cProjectionMatrix, vszStreamNames);

	iParendId = -1;
	szCameraName = "Five";
	performTest("Lookup: camera acel of a second level child (first level: property).",
		spcScene, iParendId, szCameraName, cProjectionMatrix, vszStreamNames);

	iParendId = -1;
	szCameraName = "Six";
	performTest("Lookup: camera acel of a second level property (first level: property).",
		spcScene, iParendId, szCameraName, cProjectionMatrix, vszStreamNames);

	iParendId = 37;
	szCameraName = "Seven";
	performTest("Multiple Children: Find a camera acel that has siblings",
		spcScene, iParendId, szCameraName, cProjectionMatrix, vszStreamNames);

	iParendId = 47;
	szCameraName = "Thirteen";
	performTest("Optional Parameter: No streams are set.",
		spcScene, iParendId, szCameraName, cProjectionMatrix, std::vector<std::string>());

	const std::array<float, 12> afProjectionMatrix2 = {11.0/89.0, 12.0/89.0, 13.0/89.0, 14.0/89.0,
	                                                   21.0/89.0, 22.0/89.0, 23.0/89.0, 24.0/89.0,
	                                                   31.0/89.0, 32.0/89.0, 33.0/89.0, 34.0/89.0 };
	const sra::float_buffer_t cProjectionMatrix2(12, afProjectionMatrix2.data());
	iParendId = 51;
	szCameraName = "Fourteen";
	performTest("Optional Parameter: Floats in matrix.",
		spcScene, iParendId, szCameraName, cProjectionMatrix2, vszStreamNames);

	// Error tests
	iParendId = -1;
	szCameraName = "Dummy";
	performTest("Error: Invalid parent.",
		spcScene, iParendId, szCameraName, cProjectionMatrix, vszStreamNames);

	iParendId = 1;
	szCameraName = "Dummy";
	performTest("Error: No scene graph.",
		PScene(), iParendId, szCameraName, cProjectionMatrix, vszStreamNames);

	return true;
}

int main(int argc, char **argv)
{
	std::cout << "This test is tailor-made for the WriteCameraProjectionTest. Any changes made there, will affect the results here." <<
	             "Especially the problem with the lookup of property nodes will future require changes in this test. The simple" <<
	             "test will always work." << std::endl;

	PScene spcScene;
	try {
		spcScene = sra::get_scene_representation(g_szSceneFile);
	}
	catch(const std::invalid_argument& e)
	{
		std::cerr << "Could not find scene file " << g_szSceneFile << " in the current folder." << std::endl;
		return -1;
	}
	PAcel spcGlobals = spcScene->GetGlobals();

	// Run tests
	simpleTest(spcScene, spcGlobals);
	if(true == g_bUseComplexTest)
	{
		complexTest(spcScene, spcGlobals);
	}

	return 0;
}


