#include "scene-core/sra-api/sra_scene.h"
#include "scene-core/XScene.h"
#include "scene-core/sra_implementation.h"
#include "scene-persistence/autogen/scenedb.pb.h"
#include <cstdio>
#include <fcntl.h>

#include <opencv2/opencv.hpp>

int main(void)
{
	/*
	 * A set of tests to read/write SRA streams stored in multichannel EXR files
	 * through the C++ SRA implementation.
	 *
	 * Instructions:
	 *
	 * 1.
	 * The sample data for the LoadEXRStreamTest is available at 
	 * https://www.dropbox.com/s/c5kndclq8vmxkci/laura.zip  
	 *
	 * Download and unzip it in the sample_data folder. The unzipped
	 * exr files should be in 
	 * [SCENELIB_ROOT]/sample_data/laura/
	 *
	 * 2.
	 * Create the folder 
	 * [SCENELIB_ROOT]/sra/build/dump
	 *
	 * 3.
	 * Run
	 * [SCENELIB_ROOT]/sra/build/LoadEXRStreamTest
	 * This test function is going to dump the stream frames in 
	 * [SCENELIB_ROOT]/sra/build/dump
	 *
	 */

	// 1. create a SRA scene using existing exr files
	std::cout << "-------------------------------------------------------------------------------------------\n";
	std::cout << "create SRA scene file with streams from existing sequence of exr files   \n";
	std::cout << "-------------------------------------------------------------------------------------------\n";
	{
		// create empty scene
		PScene scene = sra::make_scene_representation("./");

		// acel id generator, starts from 2 because 1 is reserved for scene globals
		sra::oid_t id = 2;

		std::string exr_seq_path = "../../sample_data/laura/%06d.exr";

		// add rgb stream
		{
			// add an empty stream called "rgb"
			PStream stream = scene->AddStream("rgb");

			// newly added streams have to be initialized, to that aim, we create an
			// acel with the necessary stream metadata
			PAcel stream_data = scene->GetGlobals()->AddChild(id++);

			{
				// path to the stream files (relative to scene working directory) - mandatory
				stream_data->AddProperty("video", id++)->SetValue()->SetString(exr_seq_path);

				// first frame - optional: if no number is specified the stream path
				// is search for files. The first frame is set according to the first
				// frame found (this process can take time, thus we provide the first
				// frame).
				stream_data->AddProperty("first_frame", id++)->SetValue()->SetInt(363015);

				// stream name - optional. This property is usefull when the string is created
				// with the XScene::AddStream(void) method.
				//	stream_data->AddProperty("name", id++)->SetValue()->SetString("rgb");

				// step between consecutive frame numbers - optional
				//	stream_data->AddProperty("frame_step", id++)->SetValue()->SetInt(1);

				// last_frame - optional 
				//	stream_data->AddProperty("frame_step", id++)->SetValue()->SetInt(363018);

				// names of the channels in the EXR files - if no channels are specified, 
				// all the channels found in the EXR file are used (as long as they
				// all have the same image dimensions and type)
				stream_data->AddProperty("exr_channel", id++);
				stream_data->Property("exr_channel")->AddChild(id++)->SetValue()->SetString("B");
				stream_data->Property("exr_channel")->AddChild(id++)->SetValue()->SetString("G");
				stream_data->Property("exr_channel")->AddChild(id++)->SetValue()->SetString("R");
			}

			std::cout << "initializing rgb stream from exr file sequence " + exr_seq_path + "\n";
			stream->Initialize(stream_data);

			// we don't need stream_data acel anymore
			scene->GetGlobals()->DeleteChildId(stream_data->Id());

			// At this point the stream is ready to use for reading/writing,
			// but this example is just about creating an scene SRA file around
			// existing EXR files. 
		}

		// add alpha stream
		{
			PStream stream = scene->AddStream("alpha");

			PAcel stream_data = scene->GetGlobals()->AddChild(id++);
			stream_data->AddProperty("video"      , id++)->SetValue()->SetString(exr_seq_path);
			stream_data->AddProperty("first_frame", id++)->SetValue()->SetInt(363015);
			stream_data->AddProperty("exr_channel", id++);
			stream_data->Property("exr_channel")->AddChild(id++)->SetValue()->SetString("A");

			std::cout << "initializing alpha stream from exr file sequence " + exr_seq_path + "\n";
			stream->Initialize(stream_data);
			scene->GetGlobals()->DeleteChildId(stream_data->Id());
		}

		// add tof depth stream
		{
			PStream stream = scene->AddStream("tof_depth");

			PAcel stream_data = scene->GetGlobals()->AddChild(id++);
			stream_data->AddProperty("video"      , id++)->SetValue()->SetString(exr_seq_path);
			stream_data->AddProperty("first_frame", id++)->SetValue()->SetInt(363015);
			stream_data->AddProperty("exr_channel", id++);
			stream_data->Property("exr_channel")->AddChild(id++)->SetValue()->SetString("depth.Z");

			std::cout << "initializing tof_depth stream from exr file sequence " + exr_seq_path + "\n";
			stream->Initialize(stream_data);
			scene->GetGlobals()->DeleteChildId(stream_data->Id());
		}

		// add tof amplitude stream
		{
			PStream stream = scene->AddStream("tof_amplitude");

			PAcel stream_data = scene->GetGlobals()->AddChild(id++);
			stream_data->AddProperty("video"      , id++)->SetValue()->SetString(exr_seq_path);
			stream_data->AddProperty("first_frame", id++)->SetValue()->SetInt(363015);
			stream_data->AddProperty("exr_channel", id++);
			stream_data->Property("exr_channel")->AddChild(id++)->SetValue()->SetString("tof.amplitude");

			stream->Initialize(stream_data);
			scene->GetGlobals()->DeleteChildId(stream_data->Id());
		}

		// add tof flags stream
		{
			PStream stream = scene->AddStream("tof_flags");

			PAcel stream_data = scene->GetGlobals()->AddChild(id++);
			stream_data->AddProperty("video"      , id++)->SetValue()->SetString(exr_seq_path);
			stream_data->AddProperty("first_frame", id++)->SetValue()->SetInt(363015);
			stream_data->AddProperty("exr_channel", id++);
			stream_data->Property("exr_channel")->AddChild(id++)->SetValue()->SetString("tof.flags");

			std::cout << "initializing tof_flags stream from exr file sequence " + exr_seq_path + "\n";
			stream->Initialize(stream_data);
			scene->GetGlobals()->DeleteChildId(stream_data->Id());
		}

		// we can also delete streams from scene. NOTE: that this does not delete
		// the stream files
		std::cout << "deleting tof_flags stream from scene\n";
		scene->DeleteStream("tof_flags");

		// we could have used a stream index (tof_flags is the 5th stream added,
		// thus its index is 4)
		//	scene->DeleteStream(4);
		
		// write the .scene file. If the third boolean parameter is true a 
		// humar-readable description of the scene file is also generated 
		// in a file named by appending .txt to the scene filename. This
		// is mainly usefull for debugging.
		std::string scene_filename = "laura.scene";
		sra::set_scene_representation(scene_filename, scene, true);

		std::cout << "output scene file: " + scene_filename + " written\n";
	}



#if 1
	// 2. load a .scene file and read streams
	std::cout << "-------------------------------------------------------------------------------------------\n";
	std::cout << "load laura.scene and read some streams \n";
	std::cout << "-------------------------------------------------------------------------------------------\n";
	{
		// load scene filename
		std::string scene_filename("./laura.scene");
		PScene scene = sra::get_scene_representation(scene_filename);

		std::cout << "Loaded " << scene_filename << ". Streams in Scene = " << scene->GetStreamCount() << std::endl;

		// read rgb stream
		{
			// get stream - returns NULLPTR if stream does not exist
			PStream stream = scene->GetStream("rgb");

			if (stream)
			{
				std::cout << "reading 'rgb' stream\n";

				while (stream->AdvanceToNextFrame()) 
				{
					PAcel frame_acel = stream->GetCurrentFrame();

					// retrieve frame info from frame_acel
					int w = frame_acel->Property(XSTREAM_INFO_VIDEO_WIDTH   )->Value()->GetInt(); // frame width
					int h = frame_acel->Property(XSTREAM_INFO_VIDEO_HEIGHT  )->Value()->GetInt(); // height
					int c = frame_acel->Property(XSTREAM_INFO_VIDEO_CHANNELS)->Value()->GetInt(); // channels
					int d = frame_acel->Property(XSTREAM_INFO_VIDEO_DEPTH   )->Value()->GetInt(); // pixel depth in bytes
					int f = frame_acel->Property(XSTREAM_INFO_VIDEO_INDEX   )->Value()->GetInt(); // frame number

					// pointer to the data
					const uint8_t* data = frame_acel->Property(XSTREAM_FRAME_VIDEO_DATA)->Value()->GetByteBuffer().Ptr();

					// NOTE: so far, the SRA only supports two OpenEXR pixel types:
					//     UINT - unsigned 16 bit integer
					//     HALF - 16 bit floating point
					//
					// EXR channels of HALF type are returned as floats (32 bits), so pixel depth
					// is either 2 (for UINT) and 4 for HALF.

					// wrap data in a opencv image and write to dump folder
					{
						int type = (d == 2) ? CV_MAKETYPE(CV_16U,c) : CV_MAKETYPE(CV_32F,c);
						cv::Mat frame(h, w, type, (void *)data);

						double min,max; cv::minMaxLoc(frame, &min, &max);

						cv::Mat frame_8u;
						frame.convertTo(frame_8u, CV_8U, 255./(max - min), -min*255./(max - min));

						char name[128];
						std::sprintf(name,"dump/rgb_%06d.png",f);
						cv::imwrite(name,frame_8u);

						std::printf("\twritten output frame dump/rgb_%06d.png\n",f);
					}
				}

				// reset stream back to initial state
				std::cout << "resetting 'rgb' stream back to initial state\n";
				stream->Reset();

				while (stream->AdvanceToNextFrame()) 
				{
					PAcel frame_acel = stream->GetCurrentFrame();

					// retrieve frame info from frame_acel
					int w = frame_acel->Property(XSTREAM_INFO_VIDEO_WIDTH   )->Value()->GetInt(); // frame width
					int h = frame_acel->Property(XSTREAM_INFO_VIDEO_HEIGHT  )->Value()->GetInt(); // height
					int c = frame_acel->Property(XSTREAM_INFO_VIDEO_CHANNELS)->Value()->GetInt(); // channels
					int d = frame_acel->Property(XSTREAM_INFO_VIDEO_DEPTH   )->Value()->GetInt(); // pixel depth in bytes
					int f = frame_acel->Property(XSTREAM_INFO_VIDEO_INDEX   )->Value()->GetInt(); // frame number

					// pointer to the data
					const uint8_t* data = frame_acel->Property(XSTREAM_FRAME_VIDEO_DATA)->Value()->GetByteBuffer().Ptr();

					// wrap data in a opencv image and write to dump folder
					{
						int type = (d == 2) ? CV_MAKETYPE(CV_16U,c) : CV_MAKETYPE(CV_32F,c);
						cv::Mat frame(h, w, type, (void *)data);

						double min,max; cv::minMaxLoc(frame, &min, &max);

						cv::Mat frame_8u;
						frame.convertTo(frame_8u, CV_8U, 255./(max - min), -min*255./(max - min));

						char name[128];
						std::sprintf(name,"dump/rgb2_%06d.png",f);
						cv::imwrite(name,frame_8u);

						std::printf("\twritten output frame dump/rgb2_%06d.png\n",f);
					}
				}
			}
			else std::cout << "No rgb stream found\n";
		}

		// read tof stream
		{
			PStream stream = scene->GetStream("tof_depth");

			if (stream)
			{
				std::cout << "reading 'tof_depth' stream\n";

				while (stream->AdvanceToNextFrame())
				{
					PAcel frame_acel = stream->GetCurrentFrame();

					// retrieve frame info from frame_acel
					int w = frame_acel->Property(XSTREAM_INFO_VIDEO_WIDTH   )->Value()->GetInt(); // frame width
					int h = frame_acel->Property(XSTREAM_INFO_VIDEO_HEIGHT  )->Value()->GetInt(); // height
					int c = frame_acel->Property(XSTREAM_INFO_VIDEO_CHANNELS)->Value()->GetInt(); // channels
					int d = frame_acel->Property(XSTREAM_INFO_VIDEO_DEPTH   )->Value()->GetInt(); // pixel depth in bytes
					int f = frame_acel->Property(XSTREAM_INFO_VIDEO_INDEX   )->Value()->GetInt(); // frame number

					// pointer to the data
					const uint8_t* data = frame_acel->Property(XSTREAM_FRAME_VIDEO_DATA)->Value()->GetByteBuffer().Ptr();

					// wrap data in a opencv image and write to dump folder
					{
						int type = (d == 2) ? CV_MAKETYPE(CV_16U,c) : CV_MAKETYPE(CV_32F,c);
						cv::Mat frame(h, w, type, (void *)data);

						double min,max; cv::minMaxLoc(frame, &min, &max);

						cv::Mat frame_8u;
						frame.convertTo(frame_8u, CV_8U, 255./(max - min), -min*255./(max - min));

						char name[128];
						std::sprintf(name,"dump/depth_%06d.png",f);
						cv::imwrite(name,frame_8u);

						std::printf("\twritten output frame dump/depth_%06d.png\n",f);
					}
				}
			}
			else std::cout << "No 'tof_depth' stream found\n";
		}
	}



	//	3. load a .scene file and add to it a new EXR stream stored in a new
	//	   exr file sequence
	std::cout << "-------------------------------------------------------------------------------------------\n";
	std::cout << "load laura.scene and add a stream stored in a separate EXR file sequence \n";
	std::cout << "-------------------------------------------------------------------------------------------\n";
	{
		// In this example, a new stream is added to laura.scene, which is just a
		// copy of the "rgb" stream. This stream will be stored in a separate
		// sequence of EXR files.

		// load scene filename
		std::cout << "loading ./laura.scene\n";
		PScene scene = sra::get_scene_representation("./laura.scene");

		// get "rgb" stream
		std::cout << "getting 'rgb' stream\n";
		PStream  in_stream = scene->GetStream("rgb");

		// add a new stream to scene
		std::cout << "adding 'rgb_cpy' stream\n";
		PStream out_stream = scene->AddStream("rgb_cpy");

		// since the newly created output stream is empty, it has to be
		// initialized
		sra::oid_t id = 2;
		PAcel stream_data = scene->GetGlobals()->AddChild(id++);

		stream_data->AddProperty("video", id++)->SetValue()->SetString("./dump/proc_%06d.exr");
		// NOTE: ./dump/proc_%06d.exr files do not exist - they will be created by the method
		// XStream::SetCurrentFrame bellow

		stream_data->AddProperty("first_frame", id++)->SetValue()->SetInt(in_stream->GetFirstFrameIndex());
		stream_data->AddProperty("frame_step" , id++)->SetValue()->SetInt(in_stream->GetFrameIndexStep());
		stream_data->AddProperty("exr_channel", id++);
		stream_data->Property("exr_channel")->AddChild(id++)->SetValue()->SetString("cpy.B");
		stream_data->Property("exr_channel")->AddChild(id++)->SetValue()->SetString("cpy.G");
		stream_data->Property("exr_channel")->AddChild(id++)->SetValue()->SetString("cpy.R");

		std::cout << "initializing 'rgb_cpy' stream to be stored in ./dump/proc_%06d.exr\n";
		out_stream->Initialize(stream_data);
		scene->GetGlobals()->DeleteChildId(stream_data->Id());

		/* BUG
		std::shared_ptr<sra::IAcelList> exr_channels(stream_data->Property("exr_channel")->List());
		std::cout << "count " << stream_data->Property("exr_channel")->List()->Count() << std::endl;
		std::cout << "count " << exr_channels->Count() << std::endl; */

		// add frames to out_stream by copying them from in_stream
		std::cout << "copying 'rgb' stream to 'rgb_cpy' stream\n";
		while (in_stream->AdvanceToNextFrame())
		{
			out_stream->AdvanceToNextFrame();
			out_stream->SetCurrentFrame(in_stream->GetCurrentFrame());
			std::cout << "\tcopied one frame\n";
		}

		// NOTE: the XStream implementation of Streams have a 'finished' status
		// which determines the behaviour of its member functions. The stream 
		// is unfinished when the current frame exist in the data (thus it can be
		// read).  After construction Streams are defined as unfinished, even if
		// they are empty.
		//
		// Stream::AdvanceToNextFrame can only be used for an unfinished stream.
		//         If called on a finished stream, it throws an exception. It
		//         advances the current frame by frame_step and loads the data,
		//         if available. If the data is available, it sets the sequence
		//         as unfinished and returns true. Otherwise, it sets the stream
		//         as finished and returns false.
		//
		// Stream::GetCurrentFrame : for an unfinished stream it returns current
		//         frame in an acel with predefined properties for frame data
		//         (width, height, data buffer, etc). For a finished stream it 
		//         returns an empty frame_acel.
		//
		// Stream::SetCurrentFrame : it writes the current frame to the stream.
		//         For an unfinished stream it will overwrite the existing
		//         current frame (or frame channel for a multimultichannel EXR
		//         stream). For a finished stream, it will create a new frame (or
		//         frame channel) at the end of the stream, increase last_frame
		//         by the frame_step and set the stream as unfinished.
		//
		// Note that once Stream::AdvanceToNextFrame reaches the end of the stream
		// (returning 'false') it cannot be called again until we call
		// Stream::SetCurrenaFrame. The reason for this is to avoid streams with
		// empty frames in between.

		// save into .scene file - create human readable representation
		sra::set_scene_representation("laura-rgb_cpy.scene",scene,true);


		std::cout << "output scene file: laura-rgb_cpy.scene written\n";
	}



	//	4. load a .scene file and add to it a new EXR stream by appending new
	//	   channels to an existing exr file sequence
	std::cout << "-------------------------------------------------------------------------------------------\n";
	std::cout << "load laura-rgb_cpy.scene and add a stream stored in an existing EXR file sequence \n";
	std::cout << "-------------------------------------------------------------------------------------------\n";
	{
		// In this example, a new stream is added to laura-rgb_cpy.scene, which
		// is just a copy of the "alpha" stream. This stream will be stored  
		// by appending new channels to an existing sequence of EXR files.

		// load scene filename
		std::cout << "loading ./laura-rgb_cpy.scene\n";
		PScene scene = sra::get_scene_representation("./laura-rgb_cpy.scene");

		// get "alpha" stream
		std::cout << "getting 'alpha' stream\n";
		PStream  in_stream = scene->GetStream("alpha");

		// add a new stream to scene
		std::cout << "adding 'alpha_cpy' stream\n";
		PStream out_stream = scene->AddStream("alpha_cpy");

		// initialize the newly created output stream
		sra::oid_t id = 2;
		PAcel stream_data = scene->GetGlobals()->AddChild(id++);

		stream_data->AddProperty("video", id++)->SetValue()->SetString("./dump/proc_%06d.exr");
		// NOTE: ./dump/proc_%06d.exr files already exist - a new channel will be
		// appended to them by the method XStream::SetCurrentFrame 

		stream_data->AddProperty("first_frame", id++)->SetValue()->SetInt(in_stream->GetFirstFrameIndex());
		stream_data->AddProperty("frame_step" , id++)->SetValue()->SetInt(in_stream->GetFrameIndexStep());
		stream_data->AddProperty("exr_channel", id++);
		stream_data->Property("exr_channel")->AddChild(id++)->SetValue()->SetString("cpy.A");

		std::cout << "initializing 'alpha_cpy' stream to be stored in ./dump/proc_%06d.exr\n";
		out_stream->Initialize(stream_data);
		scene->GetGlobals()->DeleteChildId(stream_data->Id());

		// read frames from source (with opencv), and add them to stream
		std::cout << "copying 'alpha' stream to 'alpha_cpy' stream\n";
		while (in_stream->AdvanceToNextFrame())
		{
			out_stream->AdvanceToNextFrame();
			out_stream->SetCurrentFrame(in_stream->GetCurrentFrame());
			std::cout << "\tcopied one frame\n";
		}

		sra::set_scene_representation("laura-rgba_cpy.scene",scene,true);

		std::cout << "output scene file: laura-rgba_cpy.scene written\n";
	}
#endif

	return EXIT_SUCCESS;
}
