#include "scene-core/sra-api/sra_scene.h"
#include "scene-core/XScene.h"
#include "scene-core/sra_implementation.h"
#include "scene-persistence/autogen/scenedb.pb.h"

#include <iostream>
#include <cstdio>
#include <fcntl.h>

#include <fstream>
#include <stdexcept>

#include <opencv2/opencv.hpp>

int main(void)
{
	/*
	 * A test to see if we can access data in sequences of files through
	 * SRA streams.
	 *
	 * Instructions:
	 *
	 * 1.
	 * The sample data for the LoadStreamTest is available at 
	 * https://www.dropbox.com/s/rpvr7fqjhffuz41/profesor.zip
	 *
	 * Download and unzip it in the sample_data folder. The unzipped
	 * exr files should be in 
	 * [SCENELIB_ROOT]/sample_data/profesor/
	 *
	 * 2.
	 * Run first 
	 * [SCENELIB_ROOT]/sra/build/WriteStreamTest 
	 * to create the profesor.scene file. The profesor.scene file will be 
	 * located in
	 * [SCENELIB_ROOT]/sra/build/
	 *
	 * 3. 
	 * Create the folder 
	 * [SCENELIB_ROOT]/sra/build/dump
	 *
	 * 4.
	 * Run
	 * [SCENELIB_ROOT]/sra/build/LoadStreamTest
	 * This test function is going to dump the stream frames in 
	 * [SCENELIB_ROOT]/sra/build/dump
	 *
	 */

	// load scene filename
	string scene_filename("./profesor.scene");
	// profesor.scene is generated by calling WriteStreamTest

	PScene scene = sra::get_scene_representation(scene_filename);

	// decode video frames and write them to disk
	std::cout << "XStreams in XScene = " << scene->GetStreamCount() << std::endl;

	{
		std::shared_ptr<sra::IStream> stream = scene->GetStream(std::string("rgb"));
		// We access stream with name. Alternatively, it can be accessed with a
		// stream index:
		// std::shared_ptr<sra::IStream> stream = scene->GetStream(0);

		if (stream)
		{
			std::cout << "\trgb stream found\n";

			while (stream->AdvanceToNextFrame())
			{
				PAcel frame_acel = stream->GetCurrentFrame();

				int w = frame_acel->Property(XSTREAM_INFO_VIDEO_WIDTH)
				                  ->Value()
				                  ->GetInt();

				int h = frame_acel->Property(XSTREAM_INFO_VIDEO_HEIGHT)
				                  ->Value()
				                  ->GetInt();

				int c = frame_acel->Property(XSTREAM_INFO_VIDEO_CHANNELS)
				                  ->Value()
				                  ->GetInt();

				int f = frame_acel->Property(XSTREAM_INFO_VIDEO_INDEX)
				                  ->Value()
				                  ->GetInt();

				const uint8_t* data = frame_acel->Property(XSTREAM_FRAME_VIDEO_DATA)
				                                ->Value()
				                                ->GetByteBuffer().Ptr();
				
				cv::Mat frame(h, w, CV_MAKETYPE(CV_8U,c), (void *)data);

				char name[128];
				std::sprintf(name,"dump/rgb_%03d.png",f);
				cv::imwrite(name,frame);

				std::printf("\t\twritten frame %d at dump/rgb_%03d.png\n",f,f);
			}
		}
		else
			std::cout << "No rbg stream found\n";
	}

//#if ccxc
	{
		std::shared_ptr<sra::IStream> stream = scene->GetStream(std::string("trimap"));

		if (stream)
		{
			std::cout << "\ttrimap stream found\n";

			while (stream->AdvanceToNextFrame())
			{
				PAcel frame_acel = stream->GetCurrentFrame();

				int w = frame_acel->Property(XSTREAM_INFO_VIDEO_WIDTH)
				                  ->Value()
				                  ->GetInt();

				int h = frame_acel->Property(XSTREAM_INFO_VIDEO_HEIGHT)
				                  ->Value()
				                  ->GetInt();

				int c = frame_acel->Property(XSTREAM_INFO_VIDEO_CHANNELS)
				                  ->Value()
				                  ->GetInt();

				int f = frame_acel->Property(XSTREAM_INFO_VIDEO_INDEX)
				                  ->Value()
				                  ->GetInt();

				const uint8_t* data = frame_acel->Property(XSTREAM_FRAME_VIDEO_DATA)
				                                ->Value()
				                                ->GetByteBuffer().Ptr();
				
				cv::Mat frame(h, w, CV_MAKETYPE(CV_8U,c), (void *)data);

				char name[128];
				std::sprintf(name,"dump/trimap_%03d.png",f);
				cv::imwrite(name,frame);

				std::printf("\t\twritten frame %d at dump/trimap_%03d.png\n",f,f);
			}
		}
		else
			std::cout << "No trimap stream found\n";
	}

	{
		std::shared_ptr<sra::IStream> stream = scene->GetStream(std::string("forward_oflow"));

		if (stream)
		{
			std::cout << "\tforward_oflow stream found\n";

			while (stream->AdvanceToNextFrame())
			{
				PAcel frame_acel = stream->GetCurrentFrame();

				int w = frame_acel->Property(XSTREAM_INFO_VIDEO_WIDTH)
				                  ->Value()
				                  ->GetInt();

				int h = frame_acel->Property(XSTREAM_INFO_VIDEO_HEIGHT)
				                  ->Value()
				                  ->GetInt();

				int c = frame_acel->Property(XSTREAM_INFO_VIDEO_CHANNELS)
				                  ->Value()
				                  ->GetInt();

				int f = frame_acel->Property(XSTREAM_INFO_VIDEO_INDEX)
				                  ->Value()
				                  ->GetInt();

				const uint8_t* data = frame_acel->Property(XSTREAM_FRAME_VIDEO_DATA)
				                                ->Value()
				                                ->GetByteBuffer().Ptr();
				
				cv::Mat frame(h, w, CV_MAKETYPE(CV_32F,c), (void *)data);

				// split flow into components
				cv::Mat *flow_components = new cv::Mat[2];
				cv::split(frame,flow_components);

				{
					// normalize to [0,255], convert to uchar and save
					double min,max; cv::minMaxLoc(flow_components[0], &min, &max);

					cv::Mat flow_0_8U;
					flow_components[0].convertTo(flow_0_8U, CV_8UC1, 255./(max - min), -min*255./(max - min));

					char name[128]; std::sprintf(name,"dump/flow_fw_%03d_u.png",f);
					cv::imwrite(name,flow_0_8U);
				}
				{
					// normalize to [0,255], convert to uchar and save
					double min,max; cv::minMaxLoc(flow_components[1], &min, &max);

					cv::Mat flow_0_8U;
					flow_components[1].convertTo(flow_0_8U, CV_8UC1, 255./(max - min), -255.*min/(max-min));

					char name[128]; std::sprintf(name,"dump/flow_fw_%03d_v.png",f);
					cv::imwrite(name,flow_0_8U);
				}

				std::printf("\t\twritten frame %d at dump/flow_fw_%03d_u.png\n",f,f);
				std::printf("\t\t                   dump/flow_fw_%03d_v.png\n",f);
			}
		}
		else std::cout << "No forward_flow stream found\n";
	}

	{
		std::shared_ptr<sra::IStream> stream = scene->GetStream(std::string("backward_oflow"));

		if (stream)
		{
			std::cout << "\tbackward_oflow stream found\n";

			while (stream->AdvanceToNextFrame())
			{
				PAcel frame_acel = stream->GetCurrentFrame();

				int w = frame_acel->Property(XSTREAM_INFO_VIDEO_WIDTH)
				                  ->Value()
				                  ->GetInt();

				int h = frame_acel->Property(XSTREAM_INFO_VIDEO_HEIGHT)
				                  ->Value()
				                  ->GetInt();

				int c = frame_acel->Property(XSTREAM_INFO_VIDEO_CHANNELS)
				                  ->Value()
				                  ->GetInt();

				int f = frame_acel->Property(XSTREAM_INFO_VIDEO_INDEX)
				                  ->Value()
				                  ->GetInt();

				const uint8_t* data = frame_acel->Property(XSTREAM_FRAME_VIDEO_DATA)
				                                ->Value()
				                                ->GetByteBuffer().Ptr();
				
				cv::Mat frame(h, w, CV_MAKETYPE(CV_32F,c), (void *)data);

				// split flow into components
				cv::Mat *flow_components = new cv::Mat[2];
				cv::split(frame,flow_components);

				{
					// normalize to [0,255], convert to uchar and save
					double min,max; cv::minMaxLoc(flow_components[0], &min, &max);

					cv::Mat flow_0_8U;
					flow_components[0].convertTo(flow_0_8U, CV_8UC1, 255./(max - min), -min*255./(max - min));

					char name[128]; std::sprintf(name,"dump/flow_bw_%03d_u.png",f);
					cv::imwrite(name,flow_0_8U);
				}
				{
					// normalize to [0,255], convert to uchar and save
					double min,max; cv::minMaxLoc(flow_components[1], &min, &max);

					cv::Mat flow_0_8U;
					flow_components[1].convertTo(flow_0_8U, CV_8UC1, 255./(max - min), -255.*min/(max-min));

					char name[128]; std::sprintf(name,"dump/flow_bw_%03d_v.png",f);
					cv::imwrite(name,flow_0_8U);
				}

				std::printf("\t\twritten frame %d at dump/flow_bw_%03d_u.png\n",f,f);
				std::printf("\t\t                   dump/flow_bw_%03d_v.png\n",f);
			}
		}
		else std::cout << "No backward_flow stream found\n";
	}
//#endif

//	sra::set_scene_representation("./profesor_after-load.scene", scene, true);

	return 0;
}
