#include <memory>
#include <iostream>
#include <cmath>

#include <opencv2/opencv.hpp>

#include <scene-core/sra-api/sra_scene.h>
#include <scene-core/sra_implementation.h>
#include "accessors.h"

namespace sra { class IScene; class IAcel; }
using std::string;
using std::shared_ptr;

template <typename T> std::shared_ptr<T> make_auto_ptr(T* o) {return std::shared_ptr<T>(o); }

class AcelData
{
  std::shared_ptr<sra::IScene> sc;
public:
  std::shared_ptr<sra::IStream> stream;
  AcelData(string file);
  void UpdateStream();
  std::shared_ptr<sra::IAcel> PatchesAcel();
  std::shared_ptr<sra::IAcel> GetCameraAcel();
};

AcelData::AcelData(string file)
{
  sc = sra::get_scene_representation(file);
  if (sc->GetStreamCount() >0)
	stream = sc->GetStream(0);
}

void AcelData::UpdateStream()
{
  stream->AdvanceToNextFrame();
}

shared_ptr<sra::IAcel> AcelData::PatchesAcel()
{
  return this->sc->GetGlobals()->Property("patches");
}

std::shared_ptr<sra::IAcel> AcelData::GetCameraAcel()
{
  return sc->GetGlobals()->Property("camera");
}

class BasicTest
{

	std::shared_ptr<AcelData> data;  // data in the scene file

	unsigned short convertDepthToSMCFormat(float inputDepth);

public:

  BasicTest(string s);

  void PrintCamera();

  int GetImage(cv::Mat &image);

  int GetDepthImage(cv::Mat &image);

};

BasicTest::BasicTest( string s )
{
  data = make_auto_ptr( new AcelData(s));
  data->UpdateStream();
}

void BasicTest::PrintCamera() {

	CameraAcelAcessor cam(data->GetCameraAcel());

	std::cout << "xfov = " << cam.xfov() << std::endl;
	std::cout << "aspect_ratio = " << cam.aspect_ratio() << std::endl;
	std::cout << "znear = " << cam.znear() << std::endl;
	std::cout << "zfar = " << cam.zfar() << std::endl;
	std::cout << "position = " << cam.position()[0] << " " << cam.position()[1] << " " << cam.position()[2] << std::endl;
	std::cout << "direction = " << cam.direction()[0] << " " << cam.direction()[1] << " " << cam.direction()[2] << std::endl;
	std::cout << "up = " << cam.up()[0] << " " << cam.up()[1] << " " << cam.up()[2] << std::endl;

	return;
}

/*
void BasicTest::RenderMeshesFrame( PImageTexture im, glm::mat4 mvp )
{

  auto patches = data->PatchesAcel();
  auto patches_list = patches->List();
  for (int i=0; i < patches_list->Count(); i++)
  {
    meshes->Assign(patches_list->Get(i));
    mod_render_mesh->Draw();
  }
}*/


int BasicTest::GetImage(cv::Mat &image) {

	std::shared_ptr<sra::IAcel> frame = data->stream->GetCurrentFrame();

	// there is a discrepancy of property names in sra_implementation.h and TextureRGBAcelAcessor - DOESNT work
	/*TextureRGBAcelAcessor ta(frame);
	image = cv::Mat(ta.height(), ta.width(), CV_8UC3, cv::Scalar(0));
	unsigned char* acelImage = (unsigned char*)ta.values().Ptr();
	int h = ta.height();
	int w = ta.width();*/

	unsigned char* acelImage = (unsigned char*) frame->Property(XSTREAM_FRAME_VIDEO_DATA)->Value()->GetByteBuffer().Ptr();

	int h = frame->Property(XSTREAM_INFO_VIDEO_HEIGHT)->Value()->GetInt();
	int w = frame->Property(XSTREAM_INFO_VIDEO_WIDTH)->Value()->GetInt();
	image = cv::Mat(h, w, CV_8UC3, cv::Scalar(0));

	for(int r = 0; r < h; r++) {
		for(int c = 0; c < w; c++) {
			image.at<cv::Vec3b>(r, c)[0] = acelImage[3*(r*w + c) + 2];
			image.at<cv::Vec3b>(r, c)[1] = acelImage[3*(r*w + c) + 1];
			image.at<cv::Vec3b>(r, c)[2] = acelImage[3*(r*w + c) + 0];

/*			for(int d = 0; d < 3; d++) {
				image.at<cv::Vec3b>(r, c)[d] = acelImage[3*(r*w + c) + d];
			}*/


		}
	}


	return 0;
}

unsigned short BasicTest::convertDepthToSMCFormat(float inputDepth) {

	unsigned short out = 0;
	float aux = 1.0 - (inputDepth/10.0); // to 1-0 range
	if(aux < 0.0) out = 0;
	else out = (unsigned short)(aux * 65536.0);

	return out;
}

int BasicTest::GetDepthImage(cv::Mat &image) {

	std::shared_ptr<sra::IAcel> frame = data->stream->GetCurrentFrame();

	float* acelImage = (float*) frame->Property(XSTREAM_FRAME_DEPTH_DATA)->Value()->GetFloatBuffer().Ptr();

	int h = frame->Property(XSTREAM_INFO_DEPTH_HEIGHT)->Value()->GetInt();
	int w = frame->Property(XSTREAM_INFO_DEPTH_WIDTH)->Value()->GetInt();
	image = cv::Mat(h, w, CV_16UC1, cv::Scalar(0.0));

	for(int r = 0; r < h; r++) {
		for(int c = 0; c < w; c++) {
			//std::cout << acelImage[r*w + c] << " ";
			image.at<unsigned short>(r, c) = convertDepthToSMCFormat(acelImage[r*w + c]);
		}
	}


	return 0;
}

int main() {

	BasicTest reader("../../../../trunk/sample_data/man/girl.scene");

	reader.PrintCamera();

	cv::Mat image;
	if(reader.GetImage(image) != 0) {
		std::cerr << "Cannot get the image." << std::endl;
		return -1;
	}

	cv::Mat depthImage;
	if(reader.GetDepthImage(depthImage) != 0) {
		std::cerr << "Cannot get the depth image." << std::endl;
		return -1;
	}

	if(!cv::imwrite("image.png", image)) {
		std::cerr << "Cannot save the image." << std::endl;
	}


	if(!cv::imwrite("depth.tif", depthImage)) {
		std::cerr << "Cannot save the depth image." << std::endl;
	}

	return 0;
}

