/*
 * WriteCameraProjectionTest.cpp
 *
 *  Created on: Jun 04, 2014
 *      Author: Michael Reinhardt
 */

// this repository
#include "scene-core/sra-api/sra_scene.h"
#include "scene-core/sra_implementation.h"
#include "scene-persistence/backend/IdManager.h"
#include "scene-persistence/backend/sraBackend.h"

// operating system
#include <memory>
#include <iostream>
#include <string>
#include <array>

static const bool g_bUseComplexTest = true;
static const std::string g_szSceneFile = "WriteCameraProjectionTest.scene";

// Unit test for adding a camera node to various locations inside a scene graph.

/********************************************************************************************/

bool performTest(const std::string szTestMessage,
                 PScene spcScene,
                 IdManager::IdManager& cIdGenerator,
                 const std::string szCameraName,
                 const sra::float_buffer_t cProjectionMatrix,
                 const std::vector<std::string> vszCorrespondingStreams,
                 const int iParentId)
{
	try {
		std::cout << "TEST: " << szTestMessage << std::endl;
		backend::addCameraProjectionMatrix(spcScene, cIdGenerator, szCameraName, cProjectionMatrix, vszCorrespondingStreams, iParentId);
		std::cout << "TEST SUCCEEDED" << std::endl;
		std::cout << "---------------------------------------------." << std::endl;
	}
	catch(const std::runtime_error& e) {
		std::cerr << "TEST FAILED: " << e.what() << std::endl;
		std::cout << "---------------------------------------------." << std::endl;
		return false;
	}
	return true;
}

bool writeGraph(const std::string& p_szFilename, const PScene& p_spcScene, const bool p_bAlsoWriteAsText)
{
	try {
		sra::set_scene_representation(p_szFilename, p_spcScene, p_bAlsoWriteAsText);
	}
	catch(std::invalid_argument& e) {
		std::cout << "Cannot save the scene file (" << p_szFilename << "): " << e.what() << std::endl;
		return false;
	}
	return true;
}

bool simpleTest(PScene spcScene, PAcel spcGlobals, IdManager::IdManager& cIdGenerator)
{
	const std::array<float, 12> afProjMatrix = {11, 12, 13, 14,
	                                            21, 22, 23, 24,
	                                            31, 32, 33, 34 };
	sra::float_buffer_t cProjMatrix(12, afProjMatrix.data());

	// The C++11 initializer list is not yet supported by VS. This prevents using a const vector.
	std::vector<std::string> vszStreams;
	vszStreams.push_back("The");
	vszStreams.push_back("weather");
	vszStreams.push_back("is");
	vszStreams.push_back("too good");
	vszStreams.push_back("for coding.");
	performTest("Insert: Add a camera acel to an empty scene graph.",
		spcScene, cIdGenerator, "Zero", cProjMatrix, vszStreams, spcGlobals->Id());


	return true;
}

bool complexTest(PScene spcScene, PAcel spcGlobals, IdManager::IdManager& cIdGenerator)
{
	// Generate some basic elements
	PAcel spcChild         = spcGlobals->AddChild(cIdGenerator.AllocateId());
	PAcel spcChildProperty = spcChild->AddProperty("PropOfChild", cIdGenerator.AllocateId());
	PAcel spcChildChild    = spcChild->AddChild(cIdGenerator.AllocateId());

	PAcel spcProperty         = spcGlobals->AddProperty("Prop", cIdGenerator.AllocateId());
	PAcel spcPropertyChild    = spcProperty->AddChild(cIdGenerator.AllocateId());
	PAcel spcPropertyProperty = spcProperty->AddProperty("PropOfProp", cIdGenerator.AllocateId());

	const std::array<float, 12> afProjMatrix = {11, 12, 13, 14,
	                                            21, 22, 23, 24,
	                                            31, 32, 33, 34 };
	const sra::float_buffer_t cProjMatrix(12, afProjMatrix.data());

	// The C++11 initializer list is not yet supported by VS. This prevents using a const vector.
	std::vector<std::string> vszStreams;
	vszStreams.push_back("No");
	vszStreams.push_back("Way");

	// Add camera to various positions inside the graph

	performTest("Insert: Add a camera acel to a first level child.",
		spcScene, cIdGenerator, "One", cProjMatrix, vszStreams, spcChild->Id());
	performTest("Insert: Add a camera acel to a second level property (first level: child)",
		spcScene, cIdGenerator, "Two", cProjMatrix, vszStreams, spcChildProperty->Id());
	performTest("Insert: Add a camera acel to a second level child (first level: child)",
		spcScene, cIdGenerator, "Three", cProjMatrix, vszStreams, spcChildChild->Id());

	performTest("Insert: Add a camera acel to a first level property",
		spcScene, cIdGenerator, "Four", cProjMatrix, vszStreams, spcProperty->Id());
	performTest("Insert: Add a camera acel to a second level child (first level: property)",
		spcScene, cIdGenerator, "Five", cProjMatrix, vszStreams, spcPropertyChild->Id());
	performTest("Insert: Add a camera acel to a second level property (first level: property)",
		spcScene, cIdGenerator, "Six", cProjMatrix, vszStreams, spcPropertyProperty->Id());

	// Add camera to a location that already has a camera acel.
	performTest("Insert: Add a camera acel next to an existing acel (first level child).",
		spcScene, cIdGenerator, "Seven", cProjMatrix, vszStreams, spcChild->Id());

	// Error tests
	performTest("Error: No name",
		spcScene, cIdGenerator, "", cProjMatrix, vszStreams, spcChild->Id());
	performTest("Error: Invalid projection matrix",
		spcScene, cIdGenerator, "Nine", sra::float_buffer_t(0, cProjMatrix.Ptr()), vszStreams, spcChild->Id());
	performTest("Error: Invalid projection matrix",
		spcScene, cIdGenerator, "Ten", sra::float_buffer_t(12, NULL), vszStreams, spcChild->Id());
	performTest("Error: Invalid parent id",
		spcScene, cIdGenerator, "Eleven", cProjMatrix, vszStreams, 200);
	performTest("Error: No scene graph",
		PScene(), cIdGenerator, "Eleven", cProjMatrix, vszStreams, spcChild->Id());

	// Parameter test
	performTest("Parameters: No stream",
		spcScene, cIdGenerator, "Thirteen", cProjMatrix, std::vector<std::string>(), spcChild->Id());
	const std::array<float, 12> afProjMatrix2 = {11.0/89.0, 12.0/89.0, 13.0/89.0, 14.0/89.0,
	                                             21.0/89.0, 22.0/89.0, 23.0/89.0, 24.0/89.0,
	                                             31.0/89.0, 32.0/89.0, 33.0/89.0, 34.0/89.0 };
	const sra::float_buffer_t cProjMatrix2(12, afProjMatrix2.data());
	performTest("Parameters: Floats in matrix",
		spcScene, cIdGenerator, "Fourteen", cProjMatrix2, vszStreams, spcChild->Id());
	return true;
}

int main(int argc, char **argv) {
	std::cout << "The interface (sraBackend.h) only supports Properties as leafs. All tests with Properties in the " <<
	             "path to the designated parent node will fail till this has changed." << std:: endl;

	// Generate scene graph
	PScene spcScene = sra::make_scene_representation(".");   // std::shared_ptr<sra::IScene>

	IdManager::IdManager cIdGenerator;
	cIdGenerator.MarkAsUsed(1); // starts from 2 because id 1 is reserved for globals
	PAcel spcGlobals = spcScene->GetGlobals();

	// Run tests
	simpleTest(spcScene, spcGlobals, cIdGenerator);

	if(true == g_bUseComplexTest)
	{
		complexTest(spcScene, spcGlobals, cIdGenerator);
	}

	// Write the final graph to disk
	return writeGraph(g_szSceneFile, spcScene, true);

	return 0;
}

