
#include <cstdio>
#include <fcntl.h>
#include <fstream>
#include <stdexcept>
#include <iostream>

#include <scene-core/sra-api/sra_scene.h>
#include <scene-core/sra_implementation.h>
#include "scene-persistence/autogen/scenedb.pb.h"
#include "scene-persistence/autogen/stream_fsio.pb.h"

#include <opencv2/opencv.hpp>


/* In this test we create a Scene SRA file with SRA streams using the SRA API. 
 * To that end, we load data from a source directory and write it into the
 * SRA streams. The data will be saved into the Scene folder, and a protobuf
 * description of the scene will be created by the SRA API. */
int main(void)
{
	// data is given as frame sequences, so we must specify first and last frame
	int first_frame = 1;
	int last_frame =  4;
	int frame_step =  1;

	// path to the input data root folder
	std::string input_dir = "../../sample_data/profesor/";

	// create a sra::Scene
	PScene scene = sra::make_scene_representation("./dump/");

	// id generator. Starts from 2 because id 1 is reserved for globals
	sra::oid_t id = 2;

	// add rgb stream
	{
		PStream stream = scene->AddStream("rgb");

		// initialize stream
		PAcel stream_data = scene->GetGlobals()->AddChild(id++);
		stream_data->AddProperty("video"      ,id++)->SetValue()->SetString("rgb-%03d.png");
		stream_data->AddProperty("first_frame",id++)->SetValue()->SetInt(first_frame);
		stream_data->AddProperty("frame_step" ,id++)->SetValue()->SetInt(frame_step);
		stream->Initialize(stream_data);

		// we don't need stream_data acel anymore
		scene->GetGlobals()->DeleteChildId(stream_data->Id());

		// create a frame acel
		PAcel frame = scene->GetGlobals()->AddChild(id++);
		frame->AddProperty(XSTREAM_FRAME_VIDEO_DATA   , id++);
		frame->AddProperty(XSTREAM_INFO_VIDEO_WIDTH   , id++);
		frame->AddProperty(XSTREAM_INFO_VIDEO_HEIGHT  , id++);
		frame->AddProperty(XSTREAM_INFO_VIDEO_DEPTH   , id++);
		frame->AddProperty(XSTREAM_INFO_VIDEO_CHANNELS, id++);

		// read frames from source (with opencv), and add them to stream
		for (int f = first_frame; f <= last_frame; ++f)
		{
			char frame_name[128];
			std::sprintf(frame_name, "frames/%03d.png", f);
			cv::Mat image = cv::imread(input_dir + frame_name, CV_LOAD_IMAGE_UNCHANGED);

			sra::byte_buffer_t data(image.cols * image.rows * image.elemSize(), image.data);
			frame->Property(XSTREAM_FRAME_VIDEO_DATA   )->SetValue()->SetByteBuffer(data);
			frame->Property(XSTREAM_INFO_VIDEO_WIDTH   )->SetValue()->SetInt(image.cols);
			frame->Property(XSTREAM_INFO_VIDEO_HEIGHT  )->SetValue()->SetInt(image.rows);
			frame->Property(XSTREAM_INFO_VIDEO_DEPTH   )->SetValue()->SetInt(image.elemSize1());
			frame->Property(XSTREAM_INFO_VIDEO_CHANNELS)->SetValue()->SetInt(image.channels());

			stream->AdvanceToNextFrame();
			stream->SetCurrentFrame(frame);
		}
		scene->GetGlobals()->DeleteChildId(frame->Id());
	}

	{
		PStream stream = scene->AddStream("trimap");

		// initialize stream
		PAcel stream_data = scene->GetGlobals()->AddChild(id++);
		stream_data->AddProperty("video"      ,id++)->SetValue()->SetString("trm-%03d.png");
		stream_data->AddProperty("first_frame",id++)->SetValue()->SetInt(first_frame);
		stream_data->AddProperty("frame_step" ,id++)->SetValue()->SetInt(frame_step);
		stream->Initialize(stream_data);

		// we don't need stream_data acel anymore
		scene->GetGlobals()->DeleteChildId(stream_data->Id());

		// create a frame acel
		PAcel frame = scene->GetGlobals()->AddChild(id++);
		frame->AddProperty(XSTREAM_FRAME_VIDEO_DATA   , id++);
		frame->AddProperty(XSTREAM_INFO_VIDEO_WIDTH   , id++);
		frame->AddProperty(XSTREAM_INFO_VIDEO_HEIGHT  , id++);
		frame->AddProperty(XSTREAM_INFO_VIDEO_DEPTH   , id++);
		frame->AddProperty(XSTREAM_INFO_VIDEO_CHANNELS, id++);

		// read frames from source (with opencv), and add them to stream
		for (int f = first_frame; f <= last_frame; ++f)
		{
			char frame_name[128];
			std::sprintf(frame_name, "trimap/tr.%03d.png", f);
			cv::Mat image = cv::imread(input_dir + frame_name, CV_LOAD_IMAGE_UNCHANGED);

			sra::byte_buffer_t data(image.cols * image.rows * image.elemSize(), image.data);
			frame->Property(XSTREAM_FRAME_VIDEO_DATA   )->SetValue()->SetByteBuffer(data);
			frame->Property(XSTREAM_INFO_VIDEO_WIDTH   )->SetValue()->SetInt(image.cols);
			frame->Property(XSTREAM_INFO_VIDEO_HEIGHT  )->SetValue()->SetInt(image.rows);
			frame->Property(XSTREAM_INFO_VIDEO_DEPTH   )->SetValue()->SetInt(image.elemSize1());
			frame->Property(XSTREAM_INFO_VIDEO_CHANNELS)->SetValue()->SetInt(image.channels());

			stream->AdvanceToNextFrame();
			stream->SetCurrentFrame(frame);
		}
		scene->GetGlobals()->DeleteChildId(frame->Id());
	}

	{
		PStream stream = scene->AddStream("trimap2");

		// initialize stream
		PAcel stream_data = scene->GetGlobals()->AddChild(id++);
		stream_data->AddProperty("video"      ,id++)->SetValue()->SetString("trm2-%03d.png");
		stream_data->AddProperty("first_frame",id++)->SetValue()->SetInt(first_frame);
		stream_data->AddProperty("frame_step" ,id++)->SetValue()->SetInt(frame_step);
		stream->Initialize(stream_data);

		// we don't need stream_data acel anymore
		scene->GetGlobals()->DeleteChildId(stream_data->Id());

		// create a frame acel
		PAcel frame = scene->GetGlobals()->AddChild(id++);
		frame->AddProperty(XSTREAM_FRAME_VIDEO_DATA   , id++);
		frame->AddProperty(XSTREAM_INFO_VIDEO_WIDTH   , id++);
		frame->AddProperty(XSTREAM_INFO_VIDEO_HEIGHT  , id++);
		frame->AddProperty(XSTREAM_INFO_VIDEO_DEPTH   , id++);
		frame->AddProperty(XSTREAM_INFO_VIDEO_CHANNELS, id++);

		// read frames from source (with opencv), and add them to stream
		for (int f = first_frame; f <= last_frame; ++f)
		{
			char frame_name[128];
			std::sprintf(frame_name, "trimap/tr.%03d.png", f);
			cv::Mat image = cv::imread(input_dir + frame_name, CV_LOAD_IMAGE_UNCHANGED);

			sra::byte_buffer_t data(image.cols * image.rows * image.elemSize(), image.data);
			frame->Property(XSTREAM_FRAME_VIDEO_DATA   )->SetValue()->SetByteBuffer(data);
			frame->Property(XSTREAM_INFO_VIDEO_WIDTH   )->SetValue()->SetInt(image.cols);
			frame->Property(XSTREAM_INFO_VIDEO_HEIGHT  )->SetValue()->SetInt(image.rows);
			frame->Property(XSTREAM_INFO_VIDEO_DEPTH   )->SetValue()->SetInt(image.elemSize1());
			frame->Property(XSTREAM_INFO_VIDEO_CHANNELS)->SetValue()->SetInt(image.channels());

			stream->AdvanceToNextFrame();
			stream->SetCurrentFrame(frame);
		}
		scene->GetGlobals()->DeleteChildId(frame->Id());
	}

	{
		PStream stream = scene->AddStream("trimap3");

		// initialize stream
		PAcel stream_data = scene->GetGlobals()->AddChild(id++);
		stream_data->AddProperty("video"      ,id++)->SetValue()->SetString("trm3-%03d.png");
		stream_data->AddProperty("first_frame",id++)->SetValue()->SetInt(first_frame);
		stream_data->AddProperty("frame_step" ,id++)->SetValue()->SetInt(frame_step);
		stream->Initialize(stream_data);

		// we don't need stream_data acel anymore
		scene->GetGlobals()->DeleteChildId(stream_data->Id());

		// create a frame acel
		PAcel frame = scene->GetGlobals()->AddChild(id++);
		frame->AddProperty(XSTREAM_FRAME_VIDEO_DATA   , id++);
		frame->AddProperty(XSTREAM_INFO_VIDEO_WIDTH   , id++);
		frame->AddProperty(XSTREAM_INFO_VIDEO_HEIGHT  , id++);
		frame->AddProperty(XSTREAM_INFO_VIDEO_DEPTH   , id++);
		frame->AddProperty(XSTREAM_INFO_VIDEO_CHANNELS, id++);

		// read frames from source (with opencv), and add them to stream
		for (int f = first_frame; f <= last_frame; ++f)
		{
			char frame_name[128];
			std::sprintf(frame_name, "trimap/tr.%03d.png", f);
			cv::Mat image = cv::imread(input_dir + frame_name, CV_LOAD_IMAGE_UNCHANGED);

			sra::byte_buffer_t data(image.cols * image.rows * image.elemSize(), image.data);
			frame->Property(XSTREAM_FRAME_VIDEO_DATA   )->SetValue()->SetByteBuffer(data);
			frame->Property(XSTREAM_INFO_VIDEO_WIDTH   )->SetValue()->SetInt(image.cols);
			frame->Property(XSTREAM_INFO_VIDEO_HEIGHT  )->SetValue()->SetInt(image.rows);
			frame->Property(XSTREAM_INFO_VIDEO_DEPTH   )->SetValue()->SetInt(image.elemSize1());
			frame->Property(XSTREAM_INFO_VIDEO_CHANNELS)->SetValue()->SetInt(image.channels());

			stream->AdvanceToNextFrame();
			stream->SetCurrentFrame(frame);
		}
		scene->GetGlobals()->DeleteChildId(frame->Id());
	}


	scene->DeleteStream(3);
	scene->DeleteStream("trimap");

//	scene->DeleteStream(30);
//	scene->DeleteStream("triceratops");




	sra::set_scene_representation("profesor-sra.scene",scene,true);

	/*
	// optical flow stream
	scene.add_streams();
	scene.mutable_streams(2)->set_video(work_dir + "oflow/fw_%03d_u.oflow");
	scene.mutable_streams(2)->SetExtension(fsio::name, "forward_oflow");
	scene.mutable_streams(2)->SetExtension(fsio::first_frame, first_frame);
	scene.mutable_streams(2)->SetExtension(fsio:: last_frame,  last_frame);

	// optical flow stream
	scene.add_streams();
	scene.mutable_streams(3)->set_video(work_dir + "oflow/bw_%03d_u.oflow");
	scene.mutable_streams(3)->SetExtension(fsio::name, "backward_oflow");
	scene.mutable_streams(3)->SetExtension(fsio::first_frame, first_frame);
	scene.mutable_streams(3)->SetExtension(fsio:: last_frame,  last_frame);

	// create scene file
	std::fstream bin_output("./profesor.scene", std::ios::out | std::ios::binary);
	scene.SerializeToOstream(&bin_output);
	bin_output.close();

	// create text description of scene file - it is usefull
	std::fstream txt_output("./profesor.scene.txt", std::ios::out);
	txt_output << scene.DebugString() ;
	txt_output.close();
	*/


	return 0;
}
