#include <memory>
#include <iostream>
#include <fstream>
#include <string>
#include <cmath>
#include <vector>

#include <scene-core/sra-api/sra_scene.h>
#include <scene-core/sra_implementation.h>
#include <scene-persistence/autogen/render.pb.h>
#include <scene-persistence/autogen/surface_tracking.pb.h>
#include <scene-core/XAcel.h>


namespace sra { class IScene; class IAcel; }
using namespace std;

/********************************************************************************************/

int main(int argc, char **argv) {

	PScene sc = sra::make_scene_representation(".");   // std::shared_ptr<sra::IScene>

	sra::oid_t idGenerator = 2; // starts from 2 because id 1 is reserved for globals
	PAcel globals = sc->GetGlobals();

	cout << " ============== Acel Id test ===========" << endl;
	cout << "Globals Id:" << globals->Id() << endl;

	// add acel for value tests
	PAcel acelVal = globals->AddChild(idGenerator++);
	if(!acelVal->SetId(8)) cout << "Cannot change acel Id." << endl;

	cout << " ============== Read/write of different types of acel values ===========" << endl;
	// test scalar values in an acel
	PAcel child = acelVal->AddChild(idGenerator++);
	child->SetValue()->SetFloat(50.3);
	cout << child->Value()->GetFloat() << endl;
	child->Value()->SetFloat(60.3);
	cout << child->Value()->GetFloat() << endl;
	child->Value()->SetFloat(70.3);
	cout << child->Value()->GetFloat() << endl;

	// test vector values in an acel
	child = acelVal->AddChild(idGenerator++);
	child->SetValue()->SetFloat3(sra::float3_t(1,2,3));
	sra::float3_t v = child->Value()->GetFloat3();
	cout << v.x << ", " << v.y << ", " << v.z << endl;

	// test byte array in an acel
	child = acelVal->AddChild(idGenerator++);
	unsigned char* buf = new unsigned char[3];
	buf[0] = 'a'; buf[1] = 'b'; buf[2] = 'c';
	child->SetValue()->SetByteBuffer(sra::byte_buffer_t(3, buf));
	sra::byte_buffer_t buf2 = child->Value()->GetByteBuffer();
	cout << buf2.Ptr() << endl;

	// test float array in an acel
	child = acelVal->AddChild(idGenerator++);
	float bufFloat[3] = {4.3, 3.3, 2.3};
	child->SetValue()->SetFloatBuffer(sra::float_buffer_t(3, bufFloat));
	sra::float_buffer_t bufFloat2 = child->Value()->GetFloatBuffer();
	for(int i = 0; i < bufFloat2.Size(); i++) cout << bufFloat2.Ptr()[i] << " ";
	cout << endl;

	cout << " ============== Adding child acels ===========" << endl;
	// add multiple acel children
	PAcel acelChildren = globals->AddChild(idGenerator++);
	for(int i = 0; i < 5; ++i) {
		child = acelChildren->AddChild(idGenerator++);
		child->SetValue()->SetFloat(10*i);
	}

	sra::IAcelList* children = acelChildren->List();
	for(int i = 0; i < children->Count(); ++i) {
		cout << "Id: " <<  children->Get(i)->Id() << " val: " << children->Get(i)->Value()->GetFloat() << endl;
	}

	cout << " ============== Deleting child acels ===========" << endl;
	// delete using index
	if(!acelChildren->DeleteChild(2)) cerr << "Cannot delete a child." << endl;
	// delete using Id
	if(!acelChildren->DeleteChildId(12)) cerr << "Cannot delete a child." << endl;

	for(int i = 0; i < children->Count(); ++i) {
		cout << "Id: " <<  children->Get(i)->Id() << " val: " << children->Get(i)->Value()->GetFloat() << endl;
	}


	cout << " ============== Modification of child acels ===========" << endl;
	// modification through IAcelList; access by index
	children->Get(0)->Value()->SetFloat(-1);

	// modification through IAcelList; access by id
	children->GetId(9)->Value()->SetFloat(-10);

	for(int i = 0; i < children->Count(); ++i) {
		cout << "Id: " <<  children->Get(i)->Id() << " val: " << children->Get(i)->Value()->GetFloat() << endl;
	}

	cout << " ============== Adding properties (name child acels) ===========" << endl;
	// add multiple properties
	PAcel acelProperties = globals->AddChild(idGenerator++);

	try {
		child = acelProperties->AddProperty("prop1", idGenerator++);
		child->SetValue()->SetInt(1);
		child = acelProperties->AddProperty("prop2", idGenerator++);
		child->SetValue()->SetInt(2);
		child = acelProperties->AddProperty("prop3", idGenerator++);
		child->SetValue()->SetInt(3);
		child = acelProperties->AddProperty("prop1", idGenerator++);  // invalid operation - throws exception
		child->SetValue()->SetInt(1);
	}
	catch(sra::ExistingPropertyError& e) {
		cout << "Adding already defined property." << endl;
	}

	cout << " ============== Accessing properties (name child acels) ===========" << endl;
	// accessing properties
	try {
		cout << "Id: " << acelProperties->Property("prop1")->Id() << " val: " << acelProperties->Property("prop1")->Value()->GetInt() << endl;
		cout << "Id: " << acelProperties->Property("prop2")->Id() << " val: " << acelProperties->Property("prop2")->Value()->GetInt() << endl;
		cout << "Id: " << acelProperties->Property("prop3")->Id() << " val: " << acelProperties->Property("prop3")->Value()->GetInt() << endl;
		// invalid operation - throws exception
		cout << "Id: " << acelProperties->Property("prop4")->Id() << " val: " << acelProperties->Property("prop4")->Value()->GetInt() << endl;
	}
	catch(sra::UnsupportedPropertyError& e) {
		cout << "Accessing undefined property." << endl;
	}

	cout << " ============== Adding children to properties ===========" << endl;
	// add child to property acel
	child = acelProperties->Property("prop3")->AddChild(idGenerator++);
	child->SetValue()->SetFloat(123.4);

	// access the child
	cout << "Child value: " << acelProperties->Property("prop3")->List()->Get(0)->Value()->GetFloat() << endl;

	cout << " ============== Deleting properties ===========" << endl;
	// deleting properties
	string pname = "prop3";
	if(!acelProperties->DeleteProperty(pname)) cout << "Cannot delete the property " << pname << endl;
	pname = "aaa";
	if(!acelProperties->DeleteProperty(pname)) cout << "Cannot delete the property " << pname << endl;

	try {
		cout << "Id: " << acelProperties->Property("prop1")->Id() << " val: " << acelProperties->Property("prop1")->Value()->GetInt() << endl;
		cout << "Id: " << acelProperties->Property("prop2")->Id() << " val: " << acelProperties->Property("prop2")->Value()->GetInt() << endl;
		// invalid operation - throws exception
		cout << "Id: " << acelProperties->Property("prop3")->Id() << " val: " << acelProperties->Property("prop3")->Value()->GetInt() << endl;
	}
	catch(sra::UnsupportedPropertyError& e) {
		cout << "Accessing undefined property." << endl;
	}

	cout << " ============== Modification of properties ===========" << endl;
	// change value
	acelProperties->Property("prop1")->Value()->SetInt(10);
	// introduce a child into property
	child = acelProperties->Property("prop1")->AddChild(idGenerator++);
	child->SetValue()->SetString("prop1 child");

	cout << "Id: " << acelProperties->Property("prop1")->Id() << " val: " << acelProperties->Property("prop1")->Value()->GetInt() << endl;
	cout << "Id: " << acelProperties->Property("prop1")->List()->Get(0)->Id() <<
			 " val: " << acelProperties->Property("prop1")->List()->Get(0)->Value()->GetString() << endl;

	cout << " ============== Adding child acels defined by protobuf message ===========" << endl;
	/*PAcel acelBAcel = globals->AddChild(idGenerator++, ren::Camera::descriptor() );
	acelBAcel->Property("xfov")->Value()->SetFloat(450);
	cout << "Id: " << acelBAcel->Id() << " xfov: " << acelBAcel->Property("xfov")->Value()->GetFloat() << endl;

	PAcel acelBAcel2 = globals->AddChild(idGenerator++, st::VisionCamera::descriptor() );
	acelBAcel2->Property("fx")->Value()->SetFloat(23.6);
	cout << "Id: " << acelBAcel2->Id() << " fx: " << acelBAcel2->Property("fx")->Value()->GetFloat() << endl;
	*/

	PAcel acelChildrenPB = globals->AddChild(idGenerator++);

	pb::Acel* a = new pb::Acel();
	a->mutable_ren_camera();
	PAcel acelBAcel = acelChildrenPB->AddChild(idGenerator++, a);
	acelBAcel->Property("xfov")->Value()->SetFloat(450);
	cout << "Id: " << acelBAcel->Id() << " xfov: " << acelBAcel->Property("xfov")->Value()->GetFloat() << endl;

	a = new pb::Acel();
	a->MutableExtension(st::st_camera);
	acelChildrenPB->AddChild(idGenerator++, a);
	acelChildrenPB->List()->Get(1)->Property("fx")->Value()->SetFloat(23.6);
	cout << "Id: " << acelChildrenPB->List()->Get(1)->Id() << " fx: " << acelChildrenPB->List()->Get(1)->Property("fx")->Value()->GetFloat() << endl;

	cout << " ============== Deleting child acels defined by protobuf message ===========" << endl;

	if(!acelChildrenPB->DeleteChild(0)) cerr << "Cannot delete a child." << endl;
	cout << "Id: " << acelChildrenPB->List()->Get(0)->Id() << endl;


	cout << " ============== Adding property acels defined by protobuf message ===========" << endl;
	PAcel acelPropertiesPB = globals->AddChild(idGenerator++);

	a = new pb::Acel();
	a->mutable_ren_camera();
	acelBAcel = acelPropertiesPB->AddProperty("CGcam", idGenerator++, a);
	acelBAcel->Property("xfov")->Value()->SetFloat(720);
	cout << "Id: " << acelBAcel->Id() << " xfov: " << acelBAcel->Property("xfov")->Value()->GetFloat() << endl;

	a = new pb::Acel();
	a->MutableExtension(st::st_camera);
	acelPropertiesPB->AddProperty("VIScam", idGenerator++, a);
	acelPropertiesPB->Property("VIScam")->Property("fx")->Value()->SetFloat(25.1);
	cout << "Id: " << acelPropertiesPB->Property("VIScam")->Id() << " fx: " << acelPropertiesPB->Property("VIScam")->Property("fx")->Value()->GetFloat() << endl;

	cout << " ============== Deleting property acels defined by protobuf message ===========" << endl;

	if(!acelPropertiesPB->DeleteProperty("VIScam")) cout << "Cannot delete the property " << pname << endl;

	try {
		// invalid operation - throws exception
		cout << "Id: " << acelPropertiesPB->Property("VIScam")->Id() << " fx: " << acelPropertiesPB->Property("VIScam")->Property("fx")->Value()->GetFloat() << endl;
	}
	catch(sra::UnsupportedPropertyError& e) {
		cout << "Accessing undefined property." << endl;
	}

	cout << " ============== Acel defined by protobuf message - modification of byte buffer ===========" << endl;
	a = new pb::Acel();
	a->mutable_ren_rgb();
	acelBAcel = acelPropertiesPB->AddProperty("tex", idGenerator++, a);

	buf = new unsigned char[3];
	buf[0] = 'd'; buf[1] = 'e'; buf[2] = 'f';
	acelBAcel->Property("values")->Value()->SetByteBuffer(sra::byte_buffer_t(3, buf));
	sra::byte_buffer_t bbuf = acelBAcel->Property("values")->Value()->GetByteBuffer();
	buf[1] = 'x';
	cout << "buf = " << buf << endl;
	cout << "bbuf = " << bbuf.Ptr() << endl;

	cout << " ============== Save scene file =========== " << endl;
	try {
		sra::set_scene_representation("test.scene", sc);
	}
	catch(std::invalid_argument& e) {
		cout << "Cannot save the scene file: " << e.what() << endl;
	}

	return 0;
}


